﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Windows.Threading;

namespace TestLibrary
{
    /// <summary>
    /// 스펙트럼 분석기
    /// </summary>
    [DisplayName("스펙트럼 분석기")]
    [Description("오디오 레벨과 주파수 데이터를 표시한다.")]
    [ToolboxItem(true)]
    [TemplatePart(Name = "PART_SpectrumCanvas", Type = typeof(Canvas))]     
    public class SpectrumAnalyzer : Control
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 최소 주파수 속성
        /// </summary>
        public static readonly DependencyProperty MinimumFrequencyProperty = DependencyProperty.Register
        (
            "MinimumFrequency",
            typeof(int),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                20,
                MinimumFrequencyPropertyChangedCallback,
                CoerceMinimumFrequencyPropertyCallback
            )
        );

        /// <summary>
        /// 최대 주파수 속성
        /// </summary>
        public static readonly DependencyProperty MaximumFrequencyProperty = DependencyProperty.Register
        (
            "MaximumFrequency",
            typeof(int),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                20000,
                MaximumFrequencyPropertyChangedCallback,
                CoerceMaximumFrequencyPropertyCallback
            )
        );

        /// <summary>
        /// 바 카운트 속성
        /// </summary>
        public static readonly DependencyProperty BarCountProperty = DependencyProperty.Register
        (
            "BarCount",
            typeof(int),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                32,
                BarCountPropertyChangedCallback,
                CoerceBarCountPropertyPropertyCallback
            )
        );

        /// <summary>
        /// 바 공간 속성
        /// </summary>
        public static readonly DependencyProperty BarSpacingProperty = DependencyProperty.Register
        (
            "BarSpacing",
            typeof(double),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                5.0d,
                BarSpacingPropertyChangedCallback,
                CoerceBarSpacingPropertyCallback
            )
        );

        /// <summary>
        /// 피크 낙하 지연 속성
        /// </summary>
        public static readonly DependencyProperty PeakFallDelayProperty = DependencyProperty.Register
        (
            "PeakFallDelay",
            typeof(int),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                10,
                PeakFallDelayPropertyChangedCallback,
                CoercePeakFallDelayPropertyCallback
            )
        );

        /// <summary>
        /// 주파스 스케일 (선형) 여부 속성
        /// </summary>
        public static readonly DependencyProperty IsFrequencyScaleLinearProperty = DependencyProperty.Register
        (
            "IsFrequencyScaleLinear",
            typeof(bool),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                false,
                IsFrequencyScaleLinearPropertyChangedCallback,
                CoerceIsFrequencyScaleLinearPropertyCallback
            )
        );

        /// <summary>
        /// 바 높이 스케일링 속성
        /// </summary>
        public static readonly DependencyProperty BarHeightScalingProperty = DependencyProperty.Register
        (
            "BarHeightScaling",
            typeof(BarHeightScalingStyle),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                BarHeightScalingStyle.Decibel,
                BarHeightScalingPropertyChangedCallback,
                CoerceBarHeightScalingPropertyCallback
            )
        );

        /// <summary>
        /// 피크 평균 여부
        /// </summary>
        public static readonly DependencyProperty AveragePeaksProperty = DependencyProperty.Register
        (
            "AveragePeaks",
            typeof(bool),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                false,
                AveragePeaksPropertyPropertyChangedCallback,
                CoerceAveragePeaksPropertyCallback
            )
        );

        /// <summary>
        /// 바 스타일 속성
        /// </summary>
        public static readonly DependencyProperty BarStyleProperty = DependencyProperty.Register
        (
            "BarStyle",
            typeof(Style),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                null,
                BarStylePropertyChangedCallback,
                CoerceBarStylePropertyCallback
            )
        );

        /// <summary>
        /// 피크 스타일 속성
        /// </summary>
        public static readonly DependencyProperty PeakStyleProperty = DependencyProperty.Register
        (
            "PeakStyle",
            typeof(Style),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                null,
                PeakStylePropertyChangedCallback,
                CoercePeakStylePropertyCallback
            )
        );

        /// <summary>
        /// 실제 바 너비 속성
        /// </summary>
        public static readonly DependencyProperty ActualBarWidthProperty = DependencyProperty.Register
        (
            "ActualBarWidth",
            typeof(double),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                0.0d,
                ActualBarWidthPropertyChangedCallback,
                CoerceActualBarWidthPropertyCallback
            )
        );

        /// <summary>
        /// 갱신 주기 속성
        /// </summary>
        public static readonly DependencyProperty RefreshIntervalProperty = DependencyProperty.Register
        (
            "RefreshInterval",
            typeof(int),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                DEFAULT_UPDATE_INTERVAL,
                RefreshIntervalPropertyChangedCallback,
                CoerceRefreshIntervalPropertyCallback
            )
        );

        /// <summary>
        /// FFT 복잡도 속성
        /// </summary>
        public static readonly DependencyProperty FFTComplexityProperty = DependencyProperty.Register
        (
            "FFTComplexity",
            typeof(FFTDataSize),
            typeof(SpectrumAnalyzer),
            new UIPropertyMetadata
            (
                FFTDataSize.FFT2048,
                FFTComplexityPropertyChangedCallback,
                CoerceFFTComplexityPropertyCallback
            )
        );

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 스케일 팩터 (선형)
        /// </summary>
        private const int SCALE_FACTOR_LINEAR = 9;

        /// <summary>
        /// 스케일 팩터 (제곱)
        /// </summary>
        private const int SCALE_FACTOR_SQUARE = 2;

        /// <summary>
        /// 최소 데시빌 값
        /// </summary>
        private const double MINIMUM_DB_VALUE = -90;

        /// <summary>
        /// 최대 데시빌 값
        /// </summary>
        private const double MAXIMUM_DB_VALUE = 0;

        /// <summary>
        /// 데시빌 스케일
        /// </summary>
        private const double DB_SCALE = (MAXIMUM_DB_VALUE - MINIMUM_DB_VALUE);

        /// <summary>
        /// 디폴트 업데이트 주기
        /// </summary>
        private const int DEFAULT_UPDATE_INTERVAL = 25;


        /// <summary>
        /// 애니메이션 디스패처 타이머
        /// </summary>
        private readonly DispatcherTimer animationDispatcherTimer;

        /// <summary>
        /// 스펙트럼 캔버스
        /// </summary>
        private Canvas spectrumCanvas;

        /// <summary>
        /// 스펙트럼 재생기
        /// </summary>
        private ISpectrumPlayer spectrumPlayer;

        /// <summary>
        /// 바 모양 리스트
        /// </summary>
        private readonly List<Shape> barShapeList = new List<Shape>();

        /// <summary>
        /// 피크 모양 리스트
        /// </summary>
        private readonly List<Shape> peakShapeList = new List<Shape>();

        /// <summary>
        /// 바 높이 배열
        /// </summary>
        private double[] barHeightArray;

        /// <summary>
        /// 피크 높이 배열
        /// </summary>
        private double[] peakHeightArray;

        /// <summary>
        /// 채널 데이터 배열
        /// </summary>
        private float[] channelDataArray = new float[2048];

        /// <summary>
        /// 채널 피크 데이터 배열
        /// </summary>
        private float[] channelPeakDataArray;

        /// <summary>
        /// 밴드 너비
        /// </summary>
        private double bandWidth = 1.0;

        /// <summary>
        /// 바 너비
        /// </summary>
        private double barWidth = 1;

        /// <summary>
        /// 최대 주파수 인덱스
        /// </summary>
        private int maximumFrequencyIndex = 2047;

        /// <summary>
        /// 최도 주파수 인덱스
        /// </summary>
        private int minimumFrequencyIndex;

        /// <summary>
        /// 최대 바 인덱스
        /// </summary>
        private int[] maximumBarIndex;

        /// <summary>
        /// 최대 바 로그 스케일 인덱스
        /// </summary>
        private int[] maximumBarLogScaleIndex;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 최소 주파수 - MinimumFrequency

        /// <summary>
        /// 최소 주파수
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public int MinimumFrequency
        {
            get
            {
                return (int)GetValue(MinimumFrequencyProperty);
            }
            set
            {
                SetValue(MinimumFrequencyProperty, value);
            }
        }

        #endregion
        #region 최대 주파수 - MaximumFrequency

        /// <summary>
        /// 최대 주파수
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public int MaximumFrequency
        {
            get
            {
                return (int)GetValue(MaximumFrequencyProperty);
            }
            set
            {
                SetValue(MaximumFrequencyProperty, value);
            }
        }

        #endregion
        #region 바 카운트 - BarCount

        /// <summary>
        /// 바 카운트
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public int BarCount
        {
            get
            {
                return (int)GetValue(BarCountProperty);
            }
            set
            {
                SetValue(BarCountProperty, value);
            }
        }

        #endregion
        #region 바 공간 - BarSpacing

        /// <summary>
        /// 바 공간
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public double BarSpacing
        {
            get
            {
                return (double)GetValue(BarSpacingProperty);
            }
            set
            {
                SetValue(BarSpacingProperty, value);
            }
        }

        #endregion
        #region 피크 낙하 지연 - PeakFallDelay

        /// <summary>
        /// 피크 낙하 지연
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public int PeakFallDelay
        {
            get
            {
                return (int)GetValue(PeakFallDelayProperty);
            }
            set
            {
                SetValue(PeakFallDelayProperty, value);
            }
        }

        #endregion
        #region 주파수 스케일 (선형) 여부 - IsFrequencyScaleLinear

        /// <summary>
        /// 주파수 스케일 (선형) 여부
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public bool IsFrequencyScaleLinear
        {
            get
            {
                return (bool)GetValue(IsFrequencyScaleLinearProperty);
            }
            set
            {
                SetValue(IsFrequencyScaleLinearProperty, value);
            }
        }

        #endregion
        #region 바 높이 스케일링 - BarHeightScaling

        /// <summary>
        /// 바 높이 스케일링
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public BarHeightScalingStyle BarHeightScaling
        {
            get
            {
                return (BarHeightScalingStyle)GetValue(BarHeightScalingProperty);
            }
            set
            {
                SetValue(BarHeightScalingProperty, value);
            }
        }

        #endregion
        #region 피크 평균 여부 - AveragePeaks

        /// <summary>
        /// 피크 평균 여부
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public bool AveragePeaks
        {
            get
            {
                return (bool)GetValue(AveragePeaksProperty);
            }
            set
            {
                SetValue(AveragePeaksProperty, value);
            }
        }

        #endregion
        #region 바 스타일 - BarStyle

        /// <summary>
        /// 바 스타일
        /// </summary>
        public Style BarStyle
        {
            get
            {
                return (Style)GetValue(BarStyleProperty);
            }
            set
            {
                SetValue(BarStyleProperty, value);
            }
        }

        #endregion
        #region 피크 스타일 - PeakStyle

        /// <summary>
        /// 피크 스타일
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public Style PeakStyle
        {
            get
            {
                return (Style)GetValue(PeakStyleProperty);
            }
            set
            {
                SetValue(PeakStyleProperty, value);
            }
        }

        #endregion
        #region 실제 바 너비 - ActualBarWidth

        /// <summary>
        /// 실제 바 너비
        /// </summary>
        public double ActualBarWidth
        {
            get
            {
                return (double)GetValue(ActualBarWidthProperty);
            }
            protected set
            {
                SetValue(ActualBarWidthProperty, value);
            }
        }

        #endregion
        #region 갱신 주기 - RefreshInterval

        /// <summary>
        /// 갱신 주기
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public int RefreshInterval
        {
            get
            {
                return (int)GetValue(RefreshIntervalProperty);
            }
            set
            {
                SetValue(RefreshIntervalProperty, value);
            }
        }               

        #endregion
        #region FFT 복잡도 - FFTComplexity

        /// <summary>
        /// FFT 복잡도
        /// </summary>
        [Category("SpectrumAnalyzer")]
        public FFTDataSize FFTComplexity
        {
            get
            {
                return (FFTDataSize)GetValue(FFTComplexityProperty);
            }
            set
            {
                SetValue(FFTComplexityProperty, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Static

        #region 생성자 - SpectrumAnalyzer()

        /// <summary>
        /// 생성자
        /// </summary>
        static SpectrumAnalyzer()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(SpectrumAnalyzer), new FrameworkPropertyMetadata(typeof(SpectrumAnalyzer)));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SpectrumAnalyzer()

        /// <summary>
        /// 생성자
        /// </summary>
        public SpectrumAnalyzer()
        {
            this.animationDispatcherTimer = new DispatcherTimer(DispatcherPriority.ApplicationIdle)
            {
                Interval = TimeSpan.FromMilliseconds(DEFAULT_UPDATE_INTERVAL),
            };

            this.animationDispatcherTimer.Tick += animationDispatcherTimer_Tick;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Privare

        #region 최대 주파수 속성 변경시 콜백 처리하기 - MaximumFrequencyPropertyChangedCallback(d, e)

        /// <summary>
        /// 최대 주파수 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MaximumFrequencyPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnMaximumFrequencyChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 최대 주파수 속성 강제시 콜백 처리하기 - CoerceMaximumFrequencyPropertyCallback(d, value)

        /// <summary>
        /// 최대 주파수 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">이벤트 객체</param>
        /// <param name="value">이벤트 입자</param>
        /// <returns>강제 값</returns>
        private static object CoerceMaximumFrequencyPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceMaximumFrequency((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 최소 주파수 속성 변경시 콜백 처리하기 - MinimumFrequencyPropertyChangedCallback(d, e)

        /// <summary>
        /// 최소 주파수 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MinimumFrequencyPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnMinimumFrequencyChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 최소 주파수 속성 강제시 콜백 처리하기 - CoerceMinimumFrequencyPropertyCallback(d, value)

        /// <summary>
        /// 최소 주파수 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceMinimumFrequencyPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceMinimumFrequency((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 바 카운트 속성 변경시 콜백 처리하기 - BarCountPropertyChangedCallback(d, e)

        /// <summary>
        /// 바 카운트 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void BarCountPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnBarCountChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 바 카운트 속성 강제시 콜백 처리하기 - CoerceBarCountPropertyPropertyCallback(d, value)

        /// <summary>
        /// 바 카운트 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceBarCountPropertyPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceBarCount((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 바 공간 속성 변경시 콜백 처리하기 - BarSpacingPropertyChangedCallback(d, e)

        /// <summary>
        /// 바 공간 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void BarSpacingPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnBarSpacingChanged((double)e.OldValue, (double)e.NewValue);
            }
        }

        #endregion
        #region 바 공간 속성 강제시 콜백 처리하기 - CoerceBarSpacingPropertyCallback(d, value)

        /// <summary>
        /// 바 공간 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceBarSpacingPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceBarSpacing((double)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 피크 낙하 지연 속성 변경시 콜백 처리하기 - PeakFallDelayPropertyChangedCallback(d, e)

        /// <summary>
        /// 피크 낙하 지연 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void PeakFallDelayPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnPeakFallDelayChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 피크 낙하 지연 속성 강제시 콜백 처리하기 - CoercePeakFallDelayPropertyCallback(d, value)

        /// <summary>
        /// 피크 낙하 지연 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoercePeakFallDelayPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoercePeakFallDelay((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 주파수 스케일 (선형) 여부 속성 변경시 콜백 처리하기 - IsFrequencyScaleLinearPropertyChangedCallback(d, e)

        /// <summary>
        /// 주파수 스케일 (선형) 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void IsFrequencyScaleLinearPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnIsFrequencyScaleLinearChanged((bool)e.OldValue, (bool)e.NewValue);
            }
        }

        #endregion
        #region 주파수 스케일 (선형) 여부 속성 강제시 콜백 처리하기 - CoerceIsFrequencyScaleLinearPropertyCallback(d, value)

        /// <summary>
        /// 주파수 스케일 (선형) 여부 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceIsFrequencyScaleLinearPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceIsFrequencyScaleLinear((bool)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 바 높이 스케일링 속성 변경시 콜백 처리하기 - BarHeightScalingPropertyChangedCallback(d, e)

        /// <summary>
        /// 바 높이 스케일링 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void BarHeightScalingPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnBarHeightScalingChanged((BarHeightScalingStyle)e.OldValue, (BarHeightScalingStyle)e.NewValue);
            }
        }

        #endregion
        #region 바 높이 스케일링 속성 강제시 콜백 처리하기 - CoerceBarHeightScalingPropertyCallback(d, value)

        /// <summary>
        /// 바 높이 스케일링 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceBarHeightScalingPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceBarHeightScaling((BarHeightScalingStyle)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 피크 평균 여부 속성 변경시 콜백 처리하기 - AveragePeaksPropertyPropertyChangedCallback(d, e)

        /// <summary>
        /// 피크 평균 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void AveragePeaksPropertyPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnAveragePeaksChanged((bool)e.OldValue, (bool)e.NewValue);
            }
        }

        #endregion
        #region 피크 평균 여부 속성 강제시 콜백 처리하기 - CoerceAveragePeaksPropertyCallback(d, value)

        /// <summary>
        /// 피크 평균 여부 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceAveragePeaksPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceAveragePeaks((bool)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 바 스타일 속성 변경시 콜백 처리하기 - BarStylePropertyChangedCallback(d, e)

        /// <summary>
        /// 바 스타일 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void BarStylePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnBarStyleChanged((Style)e.OldValue, (Style)e.NewValue);
            }
        }

        #endregion
        #region 바 스타일 속성 강제시 콜백 처리하기 - CoerceBarStylePropertyCallback(d, value)

        /// <summary>
        /// 바 스타일 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceBarStylePropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceBarStyle((Style)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 피크 스타일 속성 변경시 콜백 처리하기 - PeakStylePropertyChangedCallback(d, e)

        /// <summary>
        /// 피크 스타일 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void PeakStylePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnPeakStyleChanged((Style)e.OldValue, (Style)e.NewValue);
            }
        }

        #endregion
        #region 피크 스타일 속성 강제시 콜백 처리하기 - CoercePeakStylePropertyCallback(d, value)

        /// <summary>
        /// 피크 스타일 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoercePeakStylePropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoercePeakStyle((Style)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 실제 바 너비 속성 변경시 콜백 처리하기 - ActualBarWidthPropertyChangedCallback(d, e)

        /// <summary>
        /// 실제 바 너비 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ActualBarWidthPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnActualBarWidthChanged((double)e.OldValue, (double)e.NewValue);
            }
        }

        #endregion
        #region 실제 바 너비 속성 강제시 콜백 처리하기 - CoerceActualBarWidthPropertyCallback(d, value)

        /// <summary>
        /// 실제 바 너비 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceActualBarWidthPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceActualBarWidth((double)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 갱신 주기 속성 변경시 콜백 처리하기 - RefreshIntervalPropertyChangedCallback(d, e)

        /// <summary>
        /// 갱신 주기 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void RefreshIntervalPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnRefreshIntervalChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 갱신 주기 속성 강제시 콜백 처리하기 - CoerceRefreshIntervalPropertyCallback(d, value)

        /// <summary>
        /// 갱신 주기 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceRefreshIntervalPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceRefreshInterval((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region FFT 복잡도 속성 변경시 콜백 처리하기 - FFTComplexityPropertyChangedCallback(d, e)

        /// <summary>
        /// FFT 복잡도 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void FFTComplexityPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                spectrumAnalyzer.OnFFTComplexityChanged((FFTDataSize)e.OldValue, (FFTDataSize)e.NewValue);
            }
        }

        #endregion
        #region FFT 복잡도 속성 강제시 콜백 처리하기 - CoerceFFTComplexityPropertyCallback(d, value)

        /// <summary>
        /// FFT 복잡도 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceFFTComplexityPropertyCallback(DependencyObject d, object value)
        {
            SpectrumAnalyzer spectrumAnalyzer = d as SpectrumAnalyzer;

            if(spectrumAnalyzer != null)
            {
                return spectrumAnalyzer.OnCoerceFFTComplexity((FFTDataSize)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 템플리트 적용시 처리하기 - OnApplyTemplate()

        /// <summary>
        /// 템플리트 적용시 처리하기
        /// </summary>
        public override void OnApplyTemplate()
        {
            this.spectrumCanvas = GetTemplateChild("PART_SpectrumCanvas") as Canvas;

            this.spectrumCanvas.SizeChanged += spectrumCanvas_SizeChanged;

            UpdateBarLayout();
        }

        #endregion

        #region 스펙트럼 재생기 등록하기 - RegisterSpectrumPlayer(spectrumPlayer)

        /// <summary>
        /// 스펙트럼 재생기 등록하기
        /// </summary>
        /// <param name="spectrumPlayer">스펙트럼 재생기</param>
        public void RegisterSpectrumPlayer(ISpectrumPlayer spectrumPlayer)
        {
            this.spectrumPlayer = spectrumPlayer;

            spectrumPlayer.PropertyChanged += spectrumPlayer_PropertyChanged;

            UpdateBarLayout();

            this.animationDispatcherTimer.Start();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Protected

        #region 템플리트 변경시 처리하기 - OnTemplateChanged(oldTemplate, newTemplate)

        /// <summary>
        /// 템플리트 변경시 처리하기
        /// </summary>
        /// <param name="oldTemplate">기존 템플리트</param>
        /// <param name="newTemplate">신규 템플리트</param>
        protected override void OnTemplateChanged(ControlTemplate oldTemplate, ControlTemplate newTemplate)
        {
            base.OnTemplateChanged(oldTemplate, newTemplate);

            if(this.spectrumCanvas != null)            
            {
                this.spectrumCanvas.SizeChanged -= spectrumCanvas_SizeChanged;
            }
        }

        #endregion

        #region 최소 주파수 변경시 처리하기 - OnMinimumFrequencyChanged(oldValue, newValue)

        /// <summary>
        /// 최소 주파수 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnMinimumFrequencyChanged(int oldValue, int newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 최소 주파수 강제시 처리하기 - OnCoerceMinimumFrequency(value)

        /// <summary>
        /// 최소 주파수 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoerceMinimumFrequency(int value)
        {
            if(value < 0)
            {
                return value = 0;
            }

            CoerceValue(MaximumFrequencyProperty);

            return value;
        }

        #endregion

        #region 최대 주파수 변경시 처리하기 - OnMaximumFrequencyChanged(oldValue, newValue)

        /// <summary>
        /// 최대 주파수 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnMaximumFrequencyChanged(int oldValue, int newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 최대 주파수 강제시 처리하기 - OnCoerceMaximumFrequency(value)

        /// <summary>
        /// 최대 주파수 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoerceMaximumFrequency(int value)
        {
            if((int)value < MinimumFrequency)
            {
                return MinimumFrequency + 1;
            }

            return value;
        }

        #endregion

        #region 바 카운트 변경시 처리하기 - OnBarCountChanged(oldValue, newValue)

        /// <summary>
        /// 바 카운트 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnBarCountChanged(int oldValue, int newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 바 카운트 강제시 처리하기 - OnCoerceBarCount(value)

        /// <summary>
        /// 바 카운트 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoerceBarCount(int value)
        {
            value = Math.Max(value, 1);

            return value;
        }

        #endregion

        #region 바 공간 변경시 처리하기 - OnBarSpacingChanged(oldValue, newValue)

        /// <summary>
        /// 바 공간 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnBarSpacingChanged(double oldValue, double newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 바 공간 강제시 처리하기 - OnCoerceBarSpacing(value)

        /// <summary>
        /// 바 공간 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual double OnCoerceBarSpacing(double value)
        {
            value = Math.Max(value, 0);

            return value;
        }

        #endregion

        #region 피크 낙하 지연 변경시 처리하기 - OnPeakFallDelayChanged(oldValue, newValue)

        /// <summary>
        /// 피크 낙하 지연 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnPeakFallDelayChanged(int oldValue, int newValue)
        {
        }

        #endregion
        #region 피크 낙하 지연 강제시 처리하기 - OnCoercePeakFallDelay(value)

        /// <summary>
        /// 피크 낙하 지연 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoercePeakFallDelay(int value)
        {
            value = Math.Max(value, 0);

            return value;
        }

        #endregion

        #region 주파수 스케일 (선형) 여부 변경시 처리하기 - OnIsFrequencyScaleLinearChanged(oldValue, newValue)

        /// <summary>
        /// 주파수 스케일 (선형) 여부 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnIsFrequencyScaleLinearChanged(bool oldValue, bool newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 주파수 스케일 (선형) 여부 강제시 처리하기 - OnCoerceIsFrequencyScaleLinear(value)

        /// <summary>
        /// 주파수 스케일 (선형) 여부 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual bool OnCoerceIsFrequencyScaleLinear(bool value)
        {
            return value;
        }

        #endregion

        #region 바 높이 스케일링 변경시 처리하기 - OnBarHeightScalingChanged(oldValue, newValue)

        /// <summary>
        /// 바 높이 스케일링 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnBarHeightScalingChanged(BarHeightScalingStyle oldValue, BarHeightScalingStyle newValue)
        {
        }

        #endregion
        #region 바 높이 스케일링 강제시 처리하기 - OnCoerceBarHeightScaling(value)

        /// <summary>
        /// 바 높이 스케일링 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual BarHeightScalingStyle OnCoerceBarHeightScaling(BarHeightScalingStyle value)
        {
            return value;
        }

        #endregion

        #region 피크 평균 여부 변경시 처리하기 - OnAveragePeaksChanged(oldValue, newValue)

        /// <summary>
        /// 피크 평균 여부 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnAveragePeaksChanged(bool oldValue, bool newValue)
        {
        }

        #endregion
        #region 피크 평균 여부 강제시 처리하기 - OnCoerceAveragePeaks(value)

        /// <summary>
        /// 피크 평균 여부 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual bool OnCoerceAveragePeaks(bool value)
        {
            return value;
        }

        #endregion

        #region 바 스타일 변경시 처리하기 - OnBarStyleChanged(oldValue, newValue)

        /// <summary>
        /// 바 스타일 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnBarStyleChanged(Style oldValue, Style newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 바 스타일 강제시 처리하기 - OnCoerceBarStyle(value)

        /// <summary>
        /// 바 스타일 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Style OnCoerceBarStyle(Style value)
        {
            return value;
        }

        #endregion

        #region 피크 스타일 변경시 처리하기 - OnPeakStyleChanged(oldValue, newValue)

        /// <summary>
        /// 피크 스타일 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnPeakStyleChanged(Style oldValue, Style newValue)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 피크 스타일 강제시 처리하기 - OnCoercePeakStyle(value)

        /// <summary>
        /// 피크 스타일 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Style OnCoercePeakStyle(Style value)
        {

            return value;
        }

        #endregion

        #region 실제 바 너비 변경시 처리하기 - OnActualBarWidthChanged(oldValue, newValue)

        /// <summary>
        /// 실제 바 너비 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnActualBarWidthChanged(double oldValue, double newValue)
        {
        }

        #endregion
        #region 실제 바 너비 강제시 처리하기 - OnCoerceActualBarWidth(value)

        /// <summary>
        /// 실제 바 너비 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual double OnCoerceActualBarWidth(double value)
        {
            return value;
        }

        #endregion

        #region 갱신 주기 변경시 처리하기 - OnRefreshIntervalChanged(oldValue, newValue)

        /// <summary>
        /// 갱신 주기 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnRefreshIntervalChanged(int oldValue, int newValue)
        {
            this.animationDispatcherTimer.Interval = TimeSpan.FromMilliseconds(newValue);
        }

        #endregion
        #region 갱신 주기 강제시 처리하기 - OnCoerceRefreshInterval(value)

        /// <summary>
        /// 갱신 주기 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoerceRefreshInterval(int value)
        {
            value = Math.Min(1000, Math.Max(10, value));

            return value;
        }

        #endregion

        #region FFT 복잡도 변경시 처리하기 - OnFFTComplexityChanged(oldValue, newValue)

        /// <summary>
        /// FFT 복잡도 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">기존 값</param>
        /// <param name="newValue">변경 값</param>
        protected virtual void OnFFTComplexityChanged(FFTDataSize oldValue, FFTDataSize newValue)
        {
            this.channelDataArray = new float[((int)newValue / 2)];
        }

        #endregion
        #region FFT 복잡도 강제시 처리하기 - OnCoerceFFTComplexity(value)

        /// <summary>
        /// FFT 복잡도 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual FFTDataSize OnCoerceFFTComplexity(FFTDataSize value)
        {            
            return value;
        }

        #endregion

        #region 렌더러 크기 변경시 처리하기 - OnRenderSizeChanged(sizeChangedInfo)

        /// <summary>
        /// 렌더러 크기 변경시 처리하기
        /// </summary>
        /// <param name="sizeChangedInfo">크기 변경 정보</param>
        protected override void OnRenderSizeChanged(SizeChangedInfo sizeChangedInfo)
        {
            base.OnRenderSizeChanged(sizeChangedInfo);

            UpdateBarLayout();

            UpdateSpectrum();
        }

        #endregion
        #region 렌더링 처리하기 - OnRender(drawingContext)

        /// <summary>
        /// 렌더링 처리하기
        /// </summary>
        /// <param name="drawingContext">그리기 컨텍스트</param>
        protected override void OnRender(DrawingContext drawingContext)
        {
            base.OnRender(drawingContext);

            UpdateBarLayout();

            UpdateSpectrum();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Event

        #region 스펙트럼 캔버스 크기 변경시 처리하기 - spectrumCanvas_SizeChanged(sender, e)

        /// <summary>
        /// 스펙트럼 캔버스 크기 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void spectrumCanvas_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            UpdateBarLayout();
        }

        #endregion
        #region 스펙트럼 재생기 속성 변경시 처리하기 - spectrumPlayer_PropertyChanged(sender, e)

        /// <summary>
        /// 스펙트럼 재생기 속성 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void spectrumPlayer_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            switch(e.PropertyName)
            {
                case "IsPlaying" :

                    if(this.spectrumPlayer.IsPlaying && !this.animationDispatcherTimer.IsEnabled)
                    {
                        this.animationDispatcherTimer.Start();
                    }

                    break;
            }
        }

        #endregion
        #region 애니메이션 디스패처 타이머 틱 처리하기 - animationDispatcherTimer_Tick(sender, e)

        /// <summary>
        /// 애니메이션 디스패처 타이머 틱 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void animationDispatcherTimer_Tick(object sender, EventArgs e)
        {
            UpdateSpectrum();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////// Function

        #region 스펙트럼 업데이트하기 - UpdateSpectrum()

        /// <summary>
        /// 스펙트럼 업데이트하기
        /// </summary>
        private void UpdateSpectrum()
        {
            if(this.spectrumPlayer == null || this.spectrumCanvas == null || this.spectrumCanvas.RenderSize.Width < 1 || this.spectrumCanvas.RenderSize.Height < 1)
            {
                return;
            }

            if(this.spectrumPlayer.IsPlaying && !this.spectrumPlayer.SetFFTData(this.channelDataArray))
            {
                return;
            }

            UpdateSpectrumShapes();
        }

        #endregion
        #region 스펙트럼 모양 업데이트하기 - UpdateSpectrumShapes()

        /// <summary>
        /// 스펙트럼 모양 업데이트하기
        /// </summary>
        private void UpdateSpectrumShapes()
        {
            bool   isAllZero       = true;
            double fftBucketHeight = 0f;
            double barHeight       = 0f;
            double lastPeakHeight  = 0f;
            double peakYPosition   = 0f;
            double height          = this.spectrumCanvas.RenderSize.Height;
            int    barIndex        = 0;
            double peakDotHeight   = Math.Max(this.barWidth / 2.0f, 1);
            double barHeightScale  = (height - peakDotHeight);

            for(int i = this.minimumFrequencyIndex; i <= this.maximumFrequencyIndex; i++)
            {
                if(!this.spectrumPlayer.IsPlaying)
                {
                    barHeight = 0f;
                }
                else
                {
                    switch(BarHeightScaling)
                    {
                        case BarHeightScalingStyle.Decibel :

                            double dbValue = 20 * Math.Log10((double)this.channelDataArray[i]);

                            fftBucketHeight = ((dbValue - MINIMUM_DB_VALUE) / DB_SCALE) * barHeightScale;

                            break;

                        case BarHeightScalingStyle.Linear :

                            fftBucketHeight = (this.channelDataArray[i] * SCALE_FACTOR_LINEAR) * barHeightScale;

                            break;

                        case BarHeightScalingStyle.Sqrt :

                            fftBucketHeight = (((Math.Sqrt((double)this.channelDataArray[i])) * SCALE_FACTOR_SQUARE) * barHeightScale);

                            break;
                    }

                    if(barHeight < fftBucketHeight)
                    {
                        barHeight = fftBucketHeight;
                    }

                    if(barHeight < 0f)
                    {
                        barHeight = 0f;
                    }
                }

                int maximumCurrentIndex = IsFrequencyScaleLinear ? this.maximumBarIndex[barIndex] : this.maximumBarLogScaleIndex[barIndex];

                if(i == maximumCurrentIndex)
                {
                    if(barHeight > height)
                    {
                        barHeight = height;
                    }

                    if(AveragePeaks && barIndex > 0)
                    {
                        barHeight = (lastPeakHeight + barHeight) / 2;
                    }

                    peakYPosition = barHeight;

                    if(this.channelPeakDataArray[barIndex] < peakYPosition)
                    {
                        this.channelPeakDataArray[barIndex] = (float)peakYPosition;
                    }
                    else
                    {
                        this.channelPeakDataArray[barIndex] = (float)(peakYPosition + (PeakFallDelay * this.channelPeakDataArray[barIndex])) / ((float)(PeakFallDelay + 1));
                    }

                    double xPosition = BarSpacing + (this.barWidth * barIndex) + (BarSpacing * barIndex) + 1;

                    this.barShapeList[barIndex].Margin = new Thickness(xPosition, (height - 1) - barHeight, 0, 0);
                    this.barShapeList[barIndex].Height = barHeight;

                    this.peakShapeList[barIndex].Margin = new Thickness(xPosition, (height - 1) - this.channelPeakDataArray[barIndex] - peakDotHeight, 0, 0);
                    this.peakShapeList[barIndex].Height = peakDotHeight;

                    if(this.channelPeakDataArray[barIndex] > 0.05)
                    {
                        isAllZero = false;
                    }

                    lastPeakHeight = barHeight;

                    barHeight = 0f;

                    barIndex++;
                }
            }

            if(isAllZero && !this.spectrumPlayer.IsPlaying)
            {
                this.animationDispatcherTimer.Stop();
            }
        }

        #endregion
        #region 바 레이아웃 업데이트하기 - UpdateBarLayout()

        /// <summary>
        /// 바 레이아웃 업데이트하기
        /// </summary>
        private void UpdateBarLayout()
        {
            if(this.spectrumPlayer == null || this.spectrumCanvas == null)
            {
                return;
            }

            this.barWidth = Math.Max(((double)(this.spectrumCanvas.RenderSize.Width - (BarSpacing * (BarCount + 1))) / (double)BarCount), 1);

            this.maximumFrequencyIndex = Math.Min(this.spectrumPlayer.GetFFTFrequencyIndex(MaximumFrequency) + 1, 2047);
            this.minimumFrequencyIndex = Math.Min(this.spectrumPlayer.GetFFTFrequencyIndex(MinimumFrequency)    , 2047);

            this.bandWidth = Math.Max(((double)(this.maximumFrequencyIndex - this.minimumFrequencyIndex)) / this.spectrumCanvas.RenderSize.Width, 1.0);

            int actualBarCount;

            if(this.barWidth >= 1.0d)
            {
                actualBarCount = BarCount;
            }
            else
            {
                actualBarCount = Math.Max((int)((this.spectrumCanvas.RenderSize.Width - BarSpacing) / (this.barWidth + BarSpacing)), 1);
            }

            this.channelPeakDataArray = new float[actualBarCount];

            int indexCount = this.maximumFrequencyIndex - this.minimumFrequencyIndex;

            int linearIndexBucketSize = (int)Math.Round((double)indexCount / (double)actualBarCount, 0);

            List<int> maximumIndexList         = new List<int>();
            List<int> maximumLogScaleIndexList = new List<int>();

            double maximumLog = Math.Log(actualBarCount, actualBarCount);

            for(int i = 1; i < actualBarCount; i++)
            {
                maximumIndexList.Add(this.minimumFrequencyIndex + (i * linearIndexBucketSize));

                int logIndex = (int)((maximumLog - Math.Log((actualBarCount + 1) - i, (actualBarCount + 1))) * indexCount) + this.minimumFrequencyIndex;

                maximumLogScaleIndexList.Add(logIndex);
            }

            maximumIndexList.Add(this.maximumFrequencyIndex);
            maximumLogScaleIndexList.Add(this.maximumFrequencyIndex);

            this.maximumBarIndex         = maximumIndexList.ToArray();
            this.maximumBarLogScaleIndex = maximumLogScaleIndexList.ToArray();

            this.barHeightArray  = new double[actualBarCount];
            this.peakHeightArray = new double[actualBarCount];

            this.spectrumCanvas.Children.Clear();

            this.barShapeList.Clear();

            this.peakShapeList.Clear();

            double height = this.spectrumCanvas.RenderSize.Height;

            double peakDotHeight = Math.Max(this.barWidth / 2.0f, 1);

            for(int i = 0; i < actualBarCount; i++)
            {
                double xPosition = BarSpacing + (this.barWidth * i) + (BarSpacing * i) + 1;

                Rectangle barRectangle = new Rectangle()
                {
                    Margin = new Thickness(xPosition, height, 0, 0),
                    Width  = this.barWidth,
                    Height = 0,
                    Style  = BarStyle
                };

                this.barShapeList.Add(barRectangle);

                Rectangle peakRectangle = new Rectangle()
                {
                    Margin = new Thickness(xPosition, height - peakDotHeight, 0, 0),
                    Width  = this.barWidth,
                    Height = peakDotHeight,
                    Style  = PeakStyle
                };

                this.peakShapeList.Add(peakRectangle);
            }

            foreach(Shape shape in this.barShapeList)
            {
                this.spectrumCanvas.Children.Add(shape);
            }

            foreach(Shape shape in this.peakShapeList)
            {
                this.spectrumCanvas.Children.Add(shape);
            }

            ActualBarWidth = this.barWidth;
        }

        #endregion
    }
}