﻿using System;
using System.ComponentModel;
using System.Text.RegularExpressions;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;

namespace TestLibrary
{
    /// <summary>
    /// 시간 에디터
    /// </summary>
    [DisplayName("시간 에디터")]
    [Description("시간 범위를 표시하고 편집한다.")]
    [ToolboxItem(true)]    
    [TemplatePart(Name = "PART_HoursTextBox"  , Type = typeof(TextBox))]
    [TemplatePart(Name = "PART_MinutesTextBox", Type = typeof(TextBox))]
    [TemplatePart(Name = "PART_SecondsTextBox", Type = typeof(TextBox))]
    [TemplatePart(Name = "PART_SpinUpButton"  , Type = typeof(Button ))]
    [TemplatePart(Name = "PART_SpinDownButton", Type = typeof(Button ))]
    public class TimeEditor : Control
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 시간 편집 필드 - TimeEditField

        /// <summary>
        /// 시간 편집 필드
        /// </summary>
        private enum TimeEditField
        {
            /// <summary>
            /// 시
            /// </summary>
            Hour,

            /// <summary>
            /// 분
            /// </summary>
            Minute,

            /// <summary>
            /// 초
            /// </summary>
            Second
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 편집시 이벤트 - Editing

        /// <summary>
        /// 편집시 이벤트
        /// </summary>
        public event TimeEditorEventHandler Editing
        {
            add
            {
                AddHandler(EditingEvent, value);
            }
            remove
            {
                RemoveHandler(EditingEvent, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 편집시 이벤트
        /// </summary>
        public static readonly RoutedEvent EditingEvent = EventManager.RegisterRoutedEvent("Editing", RoutingStrategy.Bubble, typeof(TimeEditorEventHandler), typeof(TimeEditor));

        /// <summary>
        /// 읽기 전용 속성
        /// </summary>
        public static readonly DependencyProperty IsReadOnlyProperty = DependencyProperty.Register
        (
            "IsReadOnly",
            typeof(bool),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                false,
                IsReadOnlyPropertyChangedCallback,
                CoerceIsReadOnlyPropertyCallback
            )
        );

        /// <summary>
        /// 최소 속성
        /// </summary>
        public static readonly DependencyProperty MinimumProperty = DependencyProperty.Register
        (
            "Minimum",
            typeof(TimeSpan),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                TimeSpan.Zero,
                MinimumPropertyChangedCallback,
                CoerceMinimumPropertyCallback
            )
        );

        /// <summary>
        /// 최대 속성
        /// </summary>
        public static readonly DependencyProperty MaximumProperty = DependencyProperty.Register
        (
            "Maximum",
            typeof(TimeSpan),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                TimeSpan.MaxValue,
                MaximumPropertyChangedCallback,
                CoerceMaximumPropertyCallback
            )
        );

        /// <summary>
        /// 값 속성
        /// </summary>
        public static readonly DependencyProperty ValueProperty = DependencyProperty.Register
        (
            "Value",
            typeof(TimeSpan),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                TimeSpan.Zero,
                ValuePropertyChangedCallback,
                CoerceValuePropertyCallback
            )
        );

        /// <summary>
        /// 시 속성
        /// </summary>
        public static readonly DependencyProperty HourProperty = DependencyProperty.Register
        (
            "Hour",
            typeof(int),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                0,
                HourPropertyChangedCallback,
                CoerceHourPropertyCallback
            )
        );

        /// <summary>
        /// 분 속성
        /// </summary>
        public static readonly DependencyProperty MinuteProperty = DependencyProperty.Register
        (
            "Minute",
            typeof(int),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                0,
                MinutePropertyChangedCallback,
                CoerceMinutePropertyCallback
            )
        );

        /// <summary>
        /// 초 속성
        /// </summary>
        public static readonly DependencyProperty SecondProperty = DependencyProperty.Register
        (
            "Second",
            typeof(double),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                0.0d,
                SecondPropertyChangedCallback,
                CoerceSecondPropertyCallback
            )
        );

        /// <summary>
        /// 캐럿 브러시 속성
        /// </summary>
        public static readonly DependencyProperty CaretBrushProperty = DependencyProperty.Register
        (
            "CaretBrush",
            typeof(Brush),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                null,
                CaretBrushPropertyChangedCallback,
                CoerceCaretBrushPropertyCallback
            )
        );

        /// <summary>
        /// 시 필요 여부
        /// </summary>
        public static readonly DependencyProperty NeedsHourProperty = DependencyProperty.Register
        (
            "NeedsHour",
            typeof(bool),
            typeof(TimeEditor),
            new UIPropertyMetadata
            (
                false,
                NeedsHourPropertyChangedCallback,
                CoerceNeedsHourPropertyCallback
            )
        );

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 값 편집 여부
        /// </summary>
        private bool updatingValue;

        /// <summary>
        /// 시 텍스트 박스
        /// </summary>
        private TextBox hourTextBox;

        /// <summary>
        /// 분 텍스트 박스
        /// </summary>
        private TextBox minuteTextBox;

        /// <summary>
        /// 초 텍스트 박스
        /// </summary>
        private TextBox secondTextBox;

        /// <summary>
        /// 스핀 UP 버튼
        /// </summary>
        private ButtonBase spinUpButton;

        /// <summary>
        /// 스핀 DOWN 버튼
        /// </summary>
        private ButtonBase spinDownButton;

        /// <summary>
        /// 시 마스크
        /// </summary>
        private string hourMask;

        /// <summary>
        /// 분 마스크
        /// </summary>
        private string minuteMask;

        /// <summary>
        /// 초 마스크
        /// </summary>
        private string secondMask;

        /// <summary>
        /// 시 로드 여부
        /// </summary>
        private bool hourLoaded;

        /// <summary>
        /// 분 로드 여부
        /// </summary>
        private bool minuteLoaded;

        /// <summary>
        /// 초 로드 여부
        /// </summary>
        private bool secondLoaded;

        /// <summary>
        /// 액티브 필드
        /// </summary>
        private TimeEditField activeField = TimeEditField.Second;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 읽기 전용 여부 - IsReadOnly

        /// <summary>
        /// 읽기 전용 여부
        /// </summary>
        public bool IsReadOnly
        {
            get
            {
                return (bool)GetValue(IsReadOnlyProperty);
            }
            set
            {
                SetValue(IsReadOnlyProperty, value);
            }
        }

        #endregion
        #region 최소 - Minimum

        /// <summary>
        /// 최소
        /// </summary>
        public TimeSpan Minimum
        {
            get
            {
                return (TimeSpan)GetValue(MinimumProperty);
            }
            set
            {
                SetValue(MinimumProperty, value);
            }
        }

        #endregion
        #region 최대 - Maximum

        /// <summary>
        /// 최대
        /// </summary>
        public TimeSpan Maximum
        {
            get
            {
                return (TimeSpan)GetValue(MaximumProperty);
            }
            set
            {
                SetValue(MaximumProperty, value);
            }
        }

        #endregion
        #region 값 - Value

        /// <summary>
        /// 값
        /// </summary>
        public TimeSpan Value
        {
            get
            {
                return (TimeSpan)GetValue(ValueProperty);
            }
            set
            {
                SetValue(ValueProperty, value);
            }
        }

        #endregion
        #region 시 - Hour

        /// <summary>
        /// 시
        /// </summary>
        public int Hour
        {
            get
            {
                return (int)GetValue(HourProperty);
            }
            set
            {
                SetValue(HourProperty, value);
            }
        }

        #endregion
        #region 분 - Minute

        /// <summary>
        /// 분
        /// </summary>
        public int Minute
        {
            get
            {
                return (int)GetValue(MinuteProperty);
            }
            set
            {
                SetValue(MinuteProperty, value);
            }
        }

        #endregion
        #region 초 - Second

        /// <summary>
        /// 초
        /// </summary>
        public double Second
        {
            get
            {
                return (double)GetValue(SecondProperty);
            }
            set
            {
                SetValue(SecondProperty, value);
            }
        }

        #endregion
        #region 캐럿 브러시 - CaretBrush

        /// <summary>
        /// 캐럿 브러시
        /// </summary>
        public Brush CaretBrush
        {
            get
            {
                return (Brush)GetValue(CaretBrushProperty);
            }
            set
            {
                SetValue(CaretBrushProperty, value);
            }
        }

        #endregion
        #region 시 필요 여부 - NeedsHour

        /// <summary>
        /// 시 필요 여부
        /// </summary>
        public bool NeedsHour
        {
            get
            {
                return (bool)GetValue(NeedsHourProperty);
            }
            protected set
            {
                SetValue(NeedsHourProperty, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Static

        #region 생성자 - TimeEditor()

        /// <summary>
        /// 생성자
        /// </summary>
        static TimeEditor()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(TimeEditor), new FrameworkPropertyMetadata(typeof(TimeEditor)));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - TimeEditor()

        /// <summary>
        /// 생성자
        /// </summary>
        public TimeEditor()
        {
            UpdateMask();            
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 마스크 생성하기 - CreateMask(minimumValue, maximumValue, allowDecimals)

        /// <summary>
        /// 마스크 생성하기
        /// </summary>
        /// <param name="minimumValue">최소 값</param>
        /// <param name="maximumValue">최대 값</param>
        /// <param name="allowDecimals">십진수 허용 여부</param>
        /// <returns>마스크</returns>
        private static string CreateMask(double minimumValue, double maximumValue, bool allowDecimals)
        {
            string maskString = string.Empty;

            bool isNegative = minimumValue < 0;

            string allowNegativeString = isNegative ? @"-?" : string.Empty;

            string allowDecimalString = allowDecimals ? @"(\.\d{0,2})?" : string.Empty;

            if(maximumValue == 0 && minimumValue == 0)
            {
                maskString = string.Format("^{0}{1}{2}$", allowNegativeString, @"\d{0,2}?", allowDecimalString);
            }
            else
            {
                int minValueLength = isNegative ? ((int)minimumValue).ToString().Length : ((int)minimumValue).ToString().Length + 1;

                int numericLength = Math.Max(minValueLength, ((int)maximumValue).ToString().Length);

                maskString = string.Format("^{0}{1}{2}{3}{4}$", allowNegativeString, @"\d{0,", numericLength, @"}?", allowDecimalString);
            }

            return maskString;
        }

        #endregion

        #region 읽기 전용 여부 속성 변경시 콜백 처리하기 - IsReadOnlyPropertyChangedCallback(d, e)

        /// <summary>
        /// 읽기 전용 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void IsReadOnlyPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnIsReadOnlyChanged((bool)e.OldValue, (bool)e.NewValue);
            }
        }

        #endregion
        #region 읽기 전용 여부 속성 강제시 콜백 처리하기 - CoerceIsReadOnlyPropertyCallback(d, value)

        /// <summary>
        /// 읽기 전용 여부 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceIsReadOnlyPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;
            if(timeEditor != null)
                return timeEditor.OnCoerceIsReadOnly((bool)value);
            else
                return value;
        }

        #endregion

        #region 최소 속성 변경시 콜백 처리하기 - MinimumPropertyChangedCallback(d, e)

        /// <summary>
        /// 최소 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MinimumPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnMinimumChanged((TimeSpan)e.OldValue, (TimeSpan)e.NewValue);
            }
        }

        #endregion
        #region 최소 속성 강제시 콜백 처리하기 - CoerceMinimumPropertyCallback(d, value)

        /// <summary>
        /// 최소 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceMinimumPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceMinimum((TimeSpan)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 최대 속성 변경시 콜백 처리하기 - MaximumPropertyChangedCallback(d, e)

        /// <summary>
        /// 최대 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MaximumPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnMaximumChanged((TimeSpan)e.OldValue, (TimeSpan)e.NewValue);
            }
        }

        #endregion
        #region 최대 속성 강제시 콜백 처리하기 - CoerceMaximumPropertyCallback(d, value)

        /// <summary>
        /// 최대 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceMaximumPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceMaximum((TimeSpan)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 값 속성 변경시 콜백 처리하기 - ValuePropertyChangedCallback(d, e)

        /// <summary>
        /// 값 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ValuePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;
            if(timeEditor != null)
                timeEditor.OnValueChanged((TimeSpan)e.OldValue, (TimeSpan)e.NewValue);
        }

        #endregion
        #region 값 속성 강제시 콜백 처리하기 - CoerceValuePropertyCallback(d, value)

        /// <summary>
        /// 값 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceValuePropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceValue((TimeSpan)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 시 속성 변경시 콜백 처리하기 - HourPropertyChangedCallback(d, e)

        /// <summary>
        /// 시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void HourPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnHourChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 시 속성 강제시 콜백 처리하기 - CoerceHourPropertyCallback(d, value)

        /// <summary>
        /// 시 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceHourPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceHour((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 분 속성 변경시 콜백 처리하기 - MinutePropertyChangedCallback(d, e)

        /// <summary>
        /// 분 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MinutePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnMinuteChanged((int)e.OldValue, (int)e.NewValue);
            }
        }

        #endregion
        #region 분 속성 강제시 콜백 처리하기 - CoerceMinutePropertyCallback(d, value)

        /// <summary>
        /// 분 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceMinutePropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceMinute((int)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 초 속성 변경시 콜백 처리하기 - SecondPropertyChangedCallback(d, e)

        /// <summary>
        /// 초 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void SecondPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnSecondChanged((double)e.OldValue, (double)e.NewValue);
            }
        }

        #endregion
        #region 초 속성 강제시 콜백 처리하기 - CoerceSecondPropertyCallback(d, value)

        /// <summary>
        /// 초 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceSecondPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceSecond((double)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 캐럿 브러시 속성 변경시 콜백 처리하기 - CaretBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 캐럿 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void CaretBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnCaretBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 캐럿 브러시 속성 강제시 콜백 처리하기 - CoerceCaretBrushPropertyCallback(d, value)

        /// <summary>
        /// 캐럿 브러시 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceCaretBrushPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceCaretBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 시 필요 여부 속성 변경시 콜백 처리하기 - NeedsHourPropertyChangedCallback(d, e)

        /// <summary>
        /// 시 필요 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void NeedsHourPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                timeEditor.OnNeedsHourChanged((bool)e.OldValue, (bool)e.NewValue);
            }
        }

        #endregion
        #region 시 필요 여부 속성 강제시 콜백 처리하기 - CoerceNeedsHourPropertyCallback(d, value)

        /// <summary>
        /// 시 필요 여부 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceNeedsHourPropertyCallback(DependencyObject d, object value)
        {
            TimeEditor timeEditor = d as TimeEditor;

            if(timeEditor != null)
            {
                return timeEditor.OnCoerceNeedsHour((bool)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 템플리트 적용시 처리하기 - OnApplyTemplate()

        /// <summary>
        /// 템플리트 적용시 처리하기
        /// </summary>
        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            this.hourTextBox    = GetTemplateChild("PART_HoursTextBox"  ) as TextBox;
            this.minuteTextBox  = GetTemplateChild("PART_MinutesTextBox") as TextBox;
            this.secondTextBox  = GetTemplateChild("PART_SecondsTextBox") as TextBox;
            this.spinUpButton   = GetTemplateChild("PART_SpinUpButton"  ) as Button;
            this.spinDownButton = GetTemplateChild("PART_SpinDownButton") as Button;

            if(this.hourTextBox != null)
            {
                this.hourTextBox.GotFocus         += hourTextBox_GotFocus;
                this.hourTextBox.PreviewKeyDown   += hourTextBox_PreviewKeyDown;
                this.hourTextBox.PreviewTextInput += hourTextBox_PreviewTextInput;
                this.hourTextBox.TextChanged      += hourTextBox_TextChanged;
                this.hourTextBox.LostFocus        += hourTextBox_LostFocus;
            }

            if(this.minuteTextBox != null)
            {
                this.minuteTextBox.GotFocus         += minuteTextBox_GotFocus;
                this.minuteTextBox.PreviewKeyDown   += minuteTextBox_PreviewKeyDown;
                this.minuteTextBox.PreviewTextInput += minuteTextBox_PreviewTextInput;
                this.minuteTextBox.TextChanged      += minuteTextBox_TextChanged;
                this.minuteTextBox.LostFocus        += minuteTextBox_LostFocus;
            }

            if(this.secondTextBox != null)
            {
                this.secondTextBox.GotFocus         += secondTextBox_GotFocus;
                this.secondTextBox.PreviewKeyDown   += secondTextBox_PreviewKeyDown;
                this.secondTextBox.PreviewTextInput += secondTextBox_PreviewTextInput;
                this.secondTextBox.TextChanged      += secondTextBox_TextChanged;
                this.secondTextBox.LostFocus        += secondTextBox_LostFocus;
            }

            if(this.spinUpButton != null)
            {
                this.spinUpButton.Focusable = false;

                this.spinUpButton.Click += spinUpButton_Click;
            }
            if(this.spinDownButton != null)
            {
                this.spinDownButton.Focusable = false;

                this.spinDownButton.Click += spinDownButton_Click;
            }

            CopyValueToTextBox();

            UpdateIsReadOnly();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Protected

        #region 템플리트 변경시 처리하기 - OnTemplateChanged(oldTemplate, newTemplate)

        /// <summary>
        /// 템플리트 변경시 처리하기
        /// </summary>
        /// <param name="oldTemplate">이전 템플리트</param>
        /// <param name="newTemplate">신규 템플리트</param>
        protected override void OnTemplateChanged(ControlTemplate oldTemplate, ControlTemplate newTemplate)
        {
            base.OnTemplateChanged(oldTemplate, newTemplate);
        }
        #endregion

        #region 읽기 전용 여부 변경시 처리하기 - OnIsReadOnlyChanged(oldValue, newValue)

        /// <summary>
        /// 읽기 전용 여부 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnIsReadOnlyChanged(bool oldValue, bool newValue)
        {
            UpdateIsReadOnly();
        }

        #endregion
        #region 읽기 전용 여부 강제시 처리하기 - OnCoerceIsReadOnly(value)

        /// <summary>
        /// 읽기 전용 여부 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual bool OnCoerceIsReadOnly(bool value)
        {            
            return value;
        }

        #endregion

        #region 최소 변경시 처리하기 - OnMinimumChanged(oldValue, newValue)

        /// <summary>
        /// 최소 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnMinimumChanged(TimeSpan oldValue, TimeSpan newValue)
        {
            CoerceValue(MaximumProperty);

            CoerceValue(ValueProperty);

            UpdateMask();
        }

        #endregion
        #region 최소 강제시 처리하기 - OnCoerceMinimum(value)

        /// <summary>
        /// 최소 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual TimeSpan OnCoerceMinimum(TimeSpan value)
        {
            return value;
        }

        #endregion

        #region 최대 변경시 처리하기 - OnMaximumChanged(oldValue, newValue)

        /// <summary>
        /// 최대 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnMaximumChanged(TimeSpan oldValue, TimeSpan newValue)
        {
            UpdateMask();

            NeedsHour = (newValue >= TimeSpan.FromHours(1.0d));
        }

        #endregion
        #region 최대 강제시 처리하기 - OnCoerceMaximum(value)

        /// <summary>
        /// 최대 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual TimeSpan OnCoerceMaximum(TimeSpan value)
        {
            if(value < Minimum)
            {
                return Minimum;
            }

            return value;
        }

        #endregion

        #region 값 변경시 처리하기 - OnValueChanged(oldValue, newValue)

        /// <summary>
        /// 값 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnValueChanged(TimeSpan oldValue, TimeSpan newValue)
        {
            if(!this.updatingValue)
            {
                this.updatingValue = true;

                Hour   = newValue.Hours;
                Minute = newValue.Minutes;
                Second = newValue.Seconds + (newValue.Milliseconds / 1000.0d);

                CopyValueToTextBox();

                this.updatingValue = false;
            }
        }

        #endregion
        #region 값 강제시 처리하기 - OnCoerceValue(value)

        /// <summary>
        /// 값 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual TimeSpan OnCoerceValue(TimeSpan value)
        {
            if(value < Minimum)
            {
                return Minimum;
            }

            if(value > Maximum)
            {
                return Maximum;
            }

            return value;
        }

        #endregion

        #region 시 변경시 처리하기 - OnHourChanged(oldValue, newValue)

        /// <summary>
        /// 시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnHourChanged(int oldValue, int newValue)
        {
            if(this.updatingValue)
            {
                return;
            }

            const int MAXIMUM_HOUR = 23;

            if(newValue > MAXIMUM_HOUR)
            {
                newValue = MAXIMUM_HOUR;
            }
            
            TimeSpan value = new TimeSpan(0, newValue, Minute, (int)Second, ((int)(Second * 1000)) % 1000);

            value = TimeSpan.FromSeconds(Math.Min(Math.Max(value.TotalSeconds, Minimum.TotalSeconds), Maximum.TotalSeconds));

            Value = value;
        }

        #endregion
        #region 시 강제시 처리하기 - OnCoerceHour(value)

        /// <summary>
        /// 시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoerceHour(int value)
        {
            return value;
        }

        #endregion

        #region 분 변경시 처리하기 - OnMinuteChanged(oldValue, newValue)

        /// <summary>
        /// 분 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnMinuteChanged(int oldValue, int newValue)
        {
            if(this.updatingValue)
            {
                return;
            }

            const int MAXIMUM_MINUTE = 59;

            if(newValue > MAXIMUM_MINUTE)
            {
                newValue = MAXIMUM_MINUTE;
            }

            TimeSpan value = new TimeSpan(0, Hour, newValue, (int)Second, ((int)(Second * 1000)) % 1000);

            value = TimeSpan.FromSeconds(Math.Min(Math.Max(value.TotalSeconds, Minimum.TotalSeconds), Maximum.TotalSeconds));

            Value = value;
        }

        #endregion
        #region 분 강제시 처리하기 - OnCoerceMinute(value)

        /// <summary>
        /// 분 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual int OnCoerceMinute(int value)
        {
            return value;
        }

        #endregion

        #region 초 변경시 처리하기 - OnSecondChanged(oldValue, newValue)

        /// <summary>
        /// 초 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnSecondChanged(double oldValue, double newValue)
        {
            if(this.updatingValue)
            {
                return;
            }

            const double MAXIMUM_SECOND = 60.0d;

            if(newValue > MAXIMUM_SECOND)
            {
                newValue = MAXIMUM_SECOND - 0.01;
            }

            TimeSpan value = new TimeSpan(0, Hour, Minute, (int)newValue, ((int)(newValue * 1000)) % 1000);

            value = TimeSpan.FromSeconds(Math.Min(Math.Max(value.TotalSeconds, Minimum.TotalSeconds), Maximum.TotalSeconds));

            Value = value;
        }

        #endregion
        #region 초 강제시 처리하기 - OnCoerceSecond(value)

        /// <summary>
        /// 초 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual double OnCoerceSecond(double value)
        {
            return value;
        }

        #endregion

        #region 캐럿 브러시 변경시 처리하기 - OnCaretBrushChanged(oldValue, newValue)

        /// <summary>
        /// 캐럿 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnCaretBrushChanged(Brush oldValue, Brush newValue)
        {
            
        }

        #endregion
        #region 캐럿 브러시 강제시 처리하기 - OnCoerceCaretBrush(value)

        /// <summary>
        /// 캐럿 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceCaretBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 시 필요 여부 변경시 처리하기 - OnNeedsHourChanged(oldValue, newValue)

        /// <summary>
        /// 시 필요 여부 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnNeedsHourChanged(bool oldValue, bool newValue)
        {
            
        }

        #endregion
        #region 시 필요 여부 강제시 처리하기 - OnCoerceNeedsHour(value)

        /// <summary>
        /// 시 필요 여부 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual bool OnCoerceNeedsHour(bool value)
        {
            return value;
        }

        #endregion

        #region 포커스 획득시 처리하기 - OnGotFocus(e)

        /// <summary>
        /// 포커스 획득시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnGotFocus(RoutedEventArgs e)
        {
            switch(this.activeField)
            {
                case TimeEditField.Hour :

                    if(this.hourTextBox != null)
                    {
                        this.hourTextBox.Focus();
                    }

                    break;

                case TimeEditField.Minute :

                    if(this.minuteTextBox != null)
                    {
                        this.minuteTextBox.Focus();
                    }

                    break;

                case TimeEditField.Second :

                    if(this.secondTextBox != null)
                    {
                        this.secondTextBox.Focus();
                    }

                    break;
            }

            base.OnGotFocus(e);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Event

        #region 시 텍스트 박스 포커스 획득시 처리하기 - hourTextBox_GotFocus(sender, e)

        /// <summary>
        /// 시 텍스트 박스 포커스 획득시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hourTextBox_GotFocus(object sender, RoutedEventArgs e)
        {
            this.activeField = TimeEditField.Hour;
        }

        #endregion
        #region 시 텍스트 박스 키 DOWN PREVIEW 처리하기 - hourTextBox_PreviewKeyDown(sender, e)

        /// <summary>
        /// 시 텍스트 박스 키 DOWN PREVIEW 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hourTextBox_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            switch(e.Key)
            {
                case Key.Enter :
                case Key.Tab   :

                    if(!FireEditingEvent(TimeEditorAction.ValueEntered, TimeEditorField.Hour).Handled)
                    {
                        ValidateHour();
                    }
                    else
                    {
                        CopyValueToTextBox();
                    }

                    break;

                case Key.Up :

                    if(!FireEditingEvent(TimeEditorAction.UpPressed, TimeEditorField.Hour).Handled)
                    {
                        IncrementHour();
                    }

                    break;

                case Key.Down :

                    if(!FireEditingEvent(TimeEditorAction.DownPressed, TimeEditorField.Hour).Handled)
                    {
                        DecrementHour();
                    }

                    break;

                case Key.Right :

                    if(this.hourTextBox.CaretIndex >= this.hourTextBox.Text.Length && this.minuteTextBox != null)
                    {
                        this.minuteTextBox.Focus();

                        this.minuteTextBox.CaretIndex = 0;

                        e.Handled = true;
                    }

                    break;

                case Key.Escape :

                    if(!FireEditingEvent(TimeEditorAction.ValueEntered, TimeEditorField.Hour).Handled)
                    {
                        ValidateHour();

                        SetFocusedElement();
                    }
                    else
                    {
                        CopyValueToTextBox();
                    }

                    break;

                default :

                    break;
            }
        }

        #endregion
        #region 시 텍스트 박스 텍스트 입력 PREVIEW 처리하기 - hourTextBox_PreviewTextInput(sender, e)

        /// <summary>
        /// 시 텍스트 박스 텍스트 입력 PREVIEW 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hourTextBox_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            string text = this.hourTextBox.Text.Remove(this.hourTextBox.SelectionStart, this.hourTextBox.SelectionLength);

            text = text.Insert(this.hourTextBox.SelectionStart, e.Text);

            e.Handled = !(Regex.IsMatch(text, this.hourMask));

            base.OnPreviewTextInput(e);
        }

        #endregion
        #region 시 텍스트 박스 텍스트 변경시 처리하기 - hourTextBox_TextChanged(sender, e)

        /// <summary>
        /// 시 텍스트 박스 텍스트 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hourTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            if(string.IsNullOrEmpty(this.hourTextBox.Text) && !this.hourLoaded)
            {
                CopyValueToTextBox();

                this.hourLoaded = true;
            }
        }

        #endregion
        #region 시 텍스트 박스 포커스 상실시 처리하기 - hourTextBox_LostFocus(sender, e)

        /// <summary>
        /// 시 텍스트 박스 포커스 상실시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hourTextBox_LostFocus(object sender, RoutedEventArgs e)
        {
            ValidateHour();
        }

        #endregion

        #region 분 텍스트 박스 포커스 획득시 처리하기 - minuteTextBox_GotFocus(sender, e)

        /// <summary>
        /// 분 텍스트 박스 포커스 획득시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void minuteTextBox_GotFocus(object sender, RoutedEventArgs e)
        {
            this.activeField = TimeEditField.Minute;
        }

        #endregion
        #region 분 텍스트 박스 키 DOWN PREVIEW 처리하기 - minuteTextBox_PreviewKeyDown(sender, e)

        /// <summary>
        /// 분 텍스트 박스 키 DOWN PREVIEW 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void minuteTextBox_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            switch(e.Key)
            {
                case Key.Enter :
                case Key.Tab   :

                    if(!FireEditingEvent(TimeEditorAction.ValueEntered, TimeEditorField.Minute).Handled)
                    {
                        ValidateMinute();
                    }
                    else
                    {
                        CopyValueToTextBox();
                    }

                    break;

                case Key.Up :

                    if(!FireEditingEvent(TimeEditorAction.UpPressed, TimeEditorField.Minute).Handled)
                    {
                        IncrementMinute();
                    }

                    break;

                case Key.Down :

                    if(!FireEditingEvent(TimeEditorAction.DownPressed, TimeEditorField.Minute).Handled)
                    {
                        DecrementMinute();
                    }

                    break;

                case Key.Right :

                    if(this.minuteTextBox.CaretIndex >= this.minuteTextBox.Text.Length && this.secondTextBox != null)
                    {
                        this.secondTextBox.Focus();

                        this.secondTextBox.CaretIndex = 0;

                        e.Handled = true;
                    }

                    break;

                case Key.Left :

                    if(this.minuteTextBox.CaretIndex <= 0 && this.hourTextBox != null && this.hourTextBox.IsVisible)
                    {
                        this.hourTextBox.Focus();

                        this.hourTextBox.CaretIndex = this.hourTextBox.Text.Length;

                        e.Handled = true;
                    }

                    break;

                case Key.Escape :

                    if(!FireEditingEvent(TimeEditorAction.ValueEntered, TimeEditorField.Minute).Handled)
                    {
                        ValidateMinute();

                        SetFocusedElement();
                    }
                    else
                    {
                        CopyValueToTextBox();
                    }

                    break;

                default :

                    break;
            }
        }

        #endregion
        #region 분 텍스트 박스 텍스트 입력 PREVIEW 처리하기 - minuteTextBox_PreviewTextInput(sender, e)

        /// <summary>
        /// 분 텍스트 박스 텍스트 입력 PREVIEW 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void minuteTextBox_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            string text = this.minuteTextBox.Text.Remove(this.minuteTextBox.SelectionStart, this.minuteTextBox.SelectionLength);

            text = text.Insert(this.minuteTextBox.SelectionStart, e.Text);

            e.Handled = !(Regex.IsMatch(text, this.minuteMask));

            base.OnPreviewTextInput(e);
        }

        #endregion
        #region 분 텍스트 박스 텍스트 변경시 처리하기 - minuteTextBox_TextChanged(sender, e)

        /// <summary>
        /// 분 텍스트 박스 텍스트 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void minuteTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            if(string.IsNullOrEmpty(this.minuteTextBox.Text) && !this.minuteLoaded)
            {
                CopyValueToTextBox();

                this.minuteLoaded = true;
            }
        }

        #endregion
        #region 분 텍스트 박스 포커스 상실시 처리하기 - minuteTextBox_LostFocus(sender, e)

        /// <summary>
        /// 분 텍스트 박스 포커스 상실시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void minuteTextBox_LostFocus(object sender, RoutedEventArgs e)
        {
            ValidateMinute();
        }

        #endregion

        #region 초 텍스트 박스 포커스 획득시 처리하기 - secondTextBox_GotFocus(sender, e)

        /// <summary>
        /// 초 텍스트 박스 포커스 획득시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void secondTextBox_GotFocus(object sender, RoutedEventArgs e)
        {
            this.activeField = TimeEditField.Second;
        }

        #endregion
        #region 초 텍스트 박스 키 DOWN PREVIEW 처리하기 - secondTextBox_PreviewKeyDown(sender, e)

        /// <summary>
        /// 초 텍스트 박스 키 DOWN PREVIEW 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void secondTextBox_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            switch(e.Key)
            {
                case Key.Enter :
                case Key.Tab   :

                    if(!FireEditingEvent(TimeEditorAction.ValueEntered, TimeEditorField.Second).Handled)
                    {
                        ValidateSecond();
                    }
                    else
                    {
                        CopyValueToTextBox();
                    }

                    break;

                case Key.Up :

                    if(!FireEditingEvent(TimeEditorAction.UpPressed, TimeEditorField.Second).Handled)
                    {
                        IncrementSecond();
                    }

                    break;

                case Key.Down :

                    if(!FireEditingEvent(TimeEditorAction.DownPressed, TimeEditorField.Second).Handled)
                    {
                        DecrementSecond();
                    }

                    break;

                case Key.Left :

                    if(this.secondTextBox.CaretIndex <= 0 && this.minuteTextBox != null)
                    {
                        this.minuteTextBox.Focus();

                        this.minuteTextBox.CaretIndex = this.minuteTextBox.Text.Length;

                        e.Handled = true;
                    }

                    break;

                case Key.Escape :

                    if(!FireEditingEvent(TimeEditorAction.ValueEntered, TimeEditorField.Second).Handled)
                    {
                        ValidateSecond();

                        SetFocusedElement();
                    }
                    else
                    {
                        CopyValueToTextBox();
                    }

                    break;

                default :

                    break;
            }
        }

        #endregion
        #region 초 텍스트 박스 텍스트 입력 PREVIEW 처리하기 - secondTextBox_PreviewTextInput(sender, e)

        /// <summary>
        /// 초 텍스트 박스 텍스트 입력 PREVIEW 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void secondTextBox_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            string text = this.secondTextBox.Text.Remove(this.secondTextBox.SelectionStart, this.secondTextBox.SelectionLength);

            text = text.Insert(this.secondTextBox.SelectionStart, e.Text);

            e.Handled = !(Regex.IsMatch(text, this.secondMask));

            base.OnPreviewTextInput(e);
        }

        #endregion
        #region 초 텍스트 박스 텍스트 변경시 처리하기 - secondTextBox_TextChanged(sender, e)

        /// <summary>
        /// 초 텍스트 박스 텍스트 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void secondTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            if(string.IsNullOrEmpty(this.secondTextBox.Text) && !this.secondLoaded)
            {
                CopyValueToTextBox();

                this.secondLoaded = true;
            }
        }

        #endregion
        #region 초 텍스트 박스 포커스 상실시 처리하기 - secondTextBox_LostFocus(sender, e)

        /// <summary>
        /// 초 텍스트 박스 포커스 상실시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void secondTextBox_LostFocus(object sender, RoutedEventArgs e)
        {
            ValidateSecond();
        }

        #endregion

        #region 스핀 업 버튼 클릭시 처리하기 - spinUpButton_Click(sender, e)

        /// <summary>
        /// 스핀 업 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void spinUpButton_Click(object sender, RoutedEventArgs e)
        {
            switch(this.activeField)
            {
                case TimeEditField.Hour :

                    if(!FireEditingEvent(TimeEditorAction.UpPressed, TimeEditorField.Hour).Handled)
                    {
                        IncrementHour();
                    }

                    break;

                case TimeEditField.Minute :

                    if(!FireEditingEvent(TimeEditorAction.UpPressed, TimeEditorField.Minute).Handled)
                    {
                        IncrementMinute();
                    }

                    break;

                case TimeEditField.Second :

                    if(!FireEditingEvent(TimeEditorAction.UpPressed, TimeEditorField.Second).Handled)
                    {
                        IncrementSecond();
                    }

                    break;
            }
        }

        #endregion
        #region 스핀 다운 버튼 클릭시 처리하기 - spinDownButton_Click(sender, e)

        /// <summary>
        /// 스핀 다운 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void spinDownButton_Click(object sender, RoutedEventArgs e)
        {
            switch(this.activeField)
            {
                case TimeEditField.Hour :

                    if(!FireEditingEvent(TimeEditorAction.DownPressed, TimeEditorField.Hour).Handled)
                    {
                        DecrementHour();
                    }

                    break;

                case TimeEditField.Minute :

                    if(!FireEditingEvent(TimeEditorAction.DownPressed, TimeEditorField.Minute).Handled)
                    {
                        DecrementMinute();
                    }

                    break;

                case TimeEditField.Second :

                    if(!FireEditingEvent(TimeEditorAction.DownPressed, TimeEditorField.Second).Handled)
                    {
                        DecrementSecond();
                    }

                    break;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////// Function

        #region 마스크 업데이트하기 - UpdateMask()

        /// <summary>
        /// 마스크 업데이트하기
        /// </summary>
        private void UpdateMask()
        {
            this.hourMask   = CreateMask(0, Minimum.TotalHours                , false);
            this.minuteMask = CreateMask(0, Math.Min(60, Maximum.TotalMinutes), false);
            this.secondMask = CreateMask(0, Math.Min(60, Maximum.TotalSeconds), true );

            CoerceValue(ValueProperty);
        }

        #endregion
        #region 텍스트 박스에 값 복사하기 - CopyValueToTextBox()

        /// <summary>
        /// 텍스트 박스에 값 복사하기
        /// </summary>
        private void CopyValueToTextBox()
        {
            if(this.hourTextBox != null)
            {
                this.hourTextBox.Text = Value.Hours.ToString("00");
            }

            if(this.minuteTextBox != null)
            {
                this.minuteTextBox.Text = Value.Minutes.ToString("00");
            }

            if(this.secondTextBox != null)
            {
                this.secondTextBox.Text = (Value.Seconds + (Value.Milliseconds / 1000.0d)).ToString("00.##");
            }
        }

        #endregion
        #region 읽기 전용 여부 업데이트하기 - UpdateIsReadOnly()

        /// <summary>
        /// 읽기 전용 여부 업데이트하기
        /// </summary>
        private void UpdateIsReadOnly()
        {
            if(this.hourTextBox != null)
            {
                this.hourTextBox.IsReadOnly = IsReadOnly;
            }

            if(this.minuteTextBox != null)
            {
                this.minuteTextBox.IsReadOnly = IsReadOnly;
            }

            if(this.secondTextBox != null)
            {
                this.secondTextBox.IsReadOnly = IsReadOnly;
            }

            if(this.spinUpButton != null)
            {
                this.spinUpButton.IsEnabled = !IsReadOnly;
            }

            if(this.spinDownButton != null)
            {
                this.spinDownButton.IsEnabled = !IsReadOnly;
            }
        }

        #endregion

        #region 편집시 이벤트 발생시키기 - FireEditingEvent(action, activeField)

        /// <summary>
        /// 편집시 이벤트 발생시키기
        /// </summary>
        /// <param name="action">타임 에디터 액션</param>
        /// <param name="field">타임 에디터 필드</param>
        /// <returns></returns>
        private TimeEditorEventArgs FireEditingEvent(TimeEditorAction action, TimeEditorField field)
        {
            TimeEditorEventArgs e = new TimeEditorEventArgs(TimeEditor.EditingEvent, action, field);

            RaiseEvent(e);

            return e;
        }

        #endregion

        #region 시 검증하기 - ValidateHour()

        /// <summary>
        /// 시 검증하기
        /// </summary>
        private void ValidateHour()
        {
            if(this.hourTextBox == null)
            {
                return;
            }

            int hour = 0;

            int.TryParse(this.hourTextBox.Text, out hour);

            Hour = hour;

            CopyValueToTextBox();
        }

        #endregion
        #region 분 검증하기 - ValidateMinute()

        /// <summary>
        /// 분 검증하기
        /// </summary>
        private void ValidateMinute()
        {
            if(this.minuteTextBox == null)
            {
                return;
            }

            int minute = 0;

            int.TryParse(this.minuteTextBox.Text, out minute);

            Minute = minute;

            CopyValueToTextBox();
        }

        #endregion
        #region 초 검증하기 - ValidateSecond()

        /// <summary>
        /// 초 검증하기
        /// </summary>
        private void ValidateSecond()
        {
            if(this.secondTextBox == null)
            {
                return;
            }

            double second = 0;

            double.TryParse(this.secondTextBox.Text, out second);

            Second = second;

            CopyValueToTextBox();
        }

        #endregion

        #region 시 증가시키기 - IncrementHour()

        /// <summary>
        /// 시 증가시키기
        /// </summary>
        private void IncrementHour()
        {
            Value = Value.Add(TimeSpan.FromHours(1));

            ValidateHour();
        }

        #endregion
        #region 시 감소시키기 - DecrementHour()

        /// <summary>
        /// 시 감소시키기
        /// </summary>
        private void DecrementHour()
        {
            Value = Value.Subtract(TimeSpan.FromHours(1));

            ValidateHour();
        }

        #endregion

        #region 분 증가시키기 - IncrementMinute()

        /// <summary>
        /// 분 증가시키기
        /// </summary>
        private void IncrementMinute()
        {
            Value = Value.Add(TimeSpan.FromMinutes(1));

            ValidateMinute();
        }

        #endregion
        #region 분 감소시키기 - DecrementMinute()

        /// <summary>
        /// 분 감소시키기
        /// </summary>
        private void DecrementMinute()
        {
            Value = Value.Subtract(TimeSpan.FromMinutes(1));

            ValidateMinute();
        }

        #endregion

        #region 초 증가시키기 - IncrementSecond()

        /// <summary>
        /// 초 증가시키기
        /// </summary>
        private void IncrementSecond()
        {
            Value = Value.Add(TimeSpan.FromSeconds(1));

            ValidateSecond();
        }

        #endregion
        #region 초 감소시키기 - DecrementSecond()

        /// <summary>
        /// 초 감소시키기
        /// </summary>
        private void DecrementSecond()
        {
            Value = Value.Subtract(TimeSpan.FromSeconds(1));

            ValidateSecond();
        }

        #endregion

        #region 포커스 엘리먼트 설정하기 - SetFocusedElement()

        /// <summary>
        /// 포커스 엘리먼트 설정하기
        /// </summary>
        private void SetFocusedElement()
        {
            FrameworkElement parentFrameworkElement = Parent as FrameworkElement;

            while(parentFrameworkElement != null && parentFrameworkElement is IInputElement && !((IInputElement)parentFrameworkElement).Focusable)
            {
                parentFrameworkElement = (FrameworkElement)parentFrameworkElement.Parent;
            }

            DependencyObject scopeDependencyObject = FocusManager.GetFocusScope(this);

            FocusManager.SetFocusedElement(scopeDependencyObject, parentFrameworkElement as IInputElement);
        }

        #endregion
    }
}