﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;

namespace TestLibrary
{
    /// <summary>
    /// 웨이브폼 타임라인
    /// </summary>
    [DisplayName("웨이브폼 타임라인")]
    [Description("스테레오 웨이브폼을 표시하고 사용자가 재생 위치를 변경할 수 있게 한다.")]
    [ToolboxItem(true)]
    [TemplatePart(Name = "PART_Waveform", Type = typeof(Canvas))]
    [TemplatePart(Name = "PART_Timeline", Type = typeof(Canvas))]
    [TemplatePart(Name = "PART_Repeat"  , Type = typeof(Canvas))]
    [TemplatePart(Name = "PART_Progress", Type = typeof(Canvas))]
    public class WaveformTimeline : Control
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 왼쪽 레벨 브러시 속성
        /// </summary>
        public static readonly DependencyProperty LeftLevelBrushProperty = DependencyProperty.Register
        (
            "LeftLevelBrush",
            typeof(Brush),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                new SolidColorBrush(Colors.Blue),
                LeftLevelBrushPropertyChangedCallback,
                CoerceLeftLevelBrushPropertyCallback
            )
        );

        /// <summary>
        /// 오른쪽 레벨 브러시 속성
        /// </summary>
        public static readonly DependencyProperty RightLevelBrushProperty = DependencyProperty.Register
        (
            "RightLevelBrush",
            typeof(Brush),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                new SolidColorBrush(Colors.Red),
                RightLevelBrushPropertyChangedCallback,
                CoerceRightLevelBrushPropertyCallback
            )
        );

        /// <summary>
        /// 진행 바 브러시 속성
        /// </summary>
        public static readonly DependencyProperty ProgressBarBrushProperty = DependencyProperty.Register
        (
            "ProgressBarBrush",
            typeof(Brush),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                new SolidColorBrush(Color.FromArgb(0xCD, 0xBA, 0x00, 0xFF)),
                ProgressBarBrushPropertyChangedCallback,
                CoerceProgressBarBrushPropertyCallback
            )
        );

        /// <summary>
        /// 진행 바 두께 속성
        /// </summary>
        public static readonly DependencyProperty ProgressBarThicknessProperty = DependencyProperty.Register
        (
            "ProgressBarThickness",
            typeof(double),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                2.0d,
                ProgressBarThicknessPropertyChangedCallback,
                CoerceProgressBarThicknessPropertyCallback
            )
        );

        /// <summary>
        /// 중앙 라인 브러시 속성
        /// </summary>
        public static readonly DependencyProperty CenterLineBrushProperty = DependencyProperty.Register
        (
            "CenterLineBrush",
            typeof(Brush),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                new SolidColorBrush(Colors.Black),
                CenterLineBrushPropertyChangedCallback,
                CoerceCenterLineBrushPropertyCallback
            )
        );

        /// <summary>
        /// 중앙 라인 두께 속성
        /// </summary>
        public static readonly DependencyProperty CenterLineThicknessProperty = DependencyProperty.Register
        (
            "CenterLineThickness",
            typeof(double),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                1.0d,
                CenterLineThicknessPropertyChangedCallback,
                CoerceCenterLineThicknessPropertyCallback
            )
        );

        /// <summary>
        /// 반복 영역 브러시 속성
        /// </summary>
        public static readonly DependencyProperty RepeatRegionBrushProperty = DependencyProperty.Register
        (
            "RepeatRegionBrush",
            typeof(Brush),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                new SolidColorBrush(Color.FromArgb(0x81, 0xF6, 0xFF, 0x00)),
                RepeatRegionBrushPropertyChangedCallback,
                CoerceRepeatRegionBrushPropertyCallback
            )
        );

        /// <summary>
        /// 반복 영역 허용 여부 속성
        /// </summary>
        public static readonly DependencyProperty AllowRepeatRegionProperty = DependencyProperty.Register
        (
            "AllowRepeatRegion",
            typeof(bool),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                true,
                AllowRepeatRegionPropertyChangedCallback,
                CoerceAllowRepeatRegionPropertyCallback
            )
        );

        /// <summary>
        /// 타임라인 틱 브러시 속성
        /// </summary>
        public static readonly DependencyProperty TimelineTickBrushProperty = DependencyProperty.Register
        (
            "TimelineTickBrush",
            typeof(Brush),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                new SolidColorBrush(Colors.Black),
                TimelineTickBrushPropertyChangedCallback,
                CoerceTimelineTickBrushPropertyCallback
            )
        );

        /// <summary>
        /// 오토 스케일 웨이브폼 캐시 속성
        /// </summary>
        public static readonly DependencyProperty AutoScaleWaveformCacheProperty = DependencyProperty.Register
        (
            "AutoScaleWaveformCache",
            typeof(bool),
            typeof(WaveformTimeline),
            new UIPropertyMetadata
            (
                false,
                AutoScaleWaveformCachePropertyChangedCallback,
                CoerceAutoScaleWaveformCachePropertyCallback
            )
        );

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 마우스 이동 허용 오차
        /// </summary>
        private const int MOUSE_MOVE_TOLERANCE = 3;

        /// <summary>
        /// 지시자 삼각형 너비
        /// </summary>
        private const int INDICATOR_TRIANGLE_WIDTH = 4;

        /// <summary>
        /// 메이저 틱 높이
        /// </summary>
        private const int MAJOR_TICK_HEIGHT = 10;

        /// <summary>
        /// 마이너 틱 높이
        /// </summary>
        private const int MINOR_TICK_HEIGHT = 3;

        /// <summary>
        /// 타임스탬프 여백
        /// </summary>
        private const int TIMESTAMP_MARGIN = 5;


        /// <summary>
        /// 사운드 플레이이어
        /// </summary>
        private IWaveformPlayer soundPlayer;

        /// <summary>
        /// 웨이브폼 캔버스
        /// </summary>
        private Canvas waveformCanvas;

        /// <summary>
        /// 반복 캔버스
        /// </summary>
        private Canvas repeatCanvas;

        /// <summary>
        /// 타임라인 캔버스
        /// </summary>
        private Canvas timelineCanvas;

        /// <summary>
        /// 진행 캔버스
        /// </summary>
        private Canvas progressCanvas;

        /// <summary>
        /// 왼쪽 패스
        /// </summary>
        private readonly Path leftPath = new Path();

        /// <summary>
        /// 오른쪽 패스
        /// </summary>
        private readonly Path rightPath = new Path();

        /// <summary>
        /// 중앙 라인
        /// </summary>
        private readonly Line centerLine = new Line();

        /// <summary>
        /// 반복 영역
        /// </summary>
        private readonly Rectangle repeatRegion = new Rectangle();

        /// <summary>
        /// 진행 라인
        /// </summary>
        private readonly Line progressLine = new Line();

        /// <summary>
        /// 진행 지시자
        /// </summary>
        private readonly Path progressIndicator = new Path();

        /// <summary>
        /// 타임라인 틱 리스트
        /// </summary>
        private readonly List<Line> timeLineTickList = new List<Line>();

        /// <summary>
        /// 타임라인 배경 영역
        /// </summary>
        private readonly Rectangle timelineBackgroundRegion = new Rectangle();

        /// <summary>
        /// 타임스탬프 텍스트 블럭 리스트
        /// </summary>
        private readonly List<TextBlock> timestampTextBlockList = new List<TextBlock>();

        /// <summary>
        /// 마우스 DOWN 여부
        /// </summary>
        private bool isMouseDown;

        /// <summary>
        /// 마우스 DOWN 포인트
        /// </summary>
        private Point mouseDownPoint;

        /// <summary>
        /// 현재 포인트
        /// </summary>
        private Point currentPoint;

        /// <summary>
        /// 시작 루프 영역
        /// </summary>
        private double startLoopRegion = -1;

        /// <summary>
        /// 종료 루프 영역
        /// </summary>
        private double endLoopRegion = -1;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 왼쪽 레벨 브러시 - LeftLevelBrush

        /// <summary>
        /// 왼쪽 레벨 브러시
        /// </summary>        
        [Category("WaveformTimeline")]
        public Brush LeftLevelBrush
        {
            get
            {
                return (Brush)GetValue(LeftLevelBrushProperty);
            }
            set
            {
                SetValue(LeftLevelBrushProperty, value);
            }
        }

        #endregion
        #region 오른쪽 레벨 브러시 - RightLevelBrush

        /// <summary>
        /// 오른쪽 레벨 브러시
        /// </summary>
        [Category("WaveformTimeline")]
        public Brush RightLevelBrush
        {
            get
            {
                return (Brush)GetValue(RightLevelBrushProperty);
            }
            set
            {
                SetValue(RightLevelBrushProperty, value);
            }
        }

        #endregion
        #region 진행 바 브러시 - ProgressBarBrush

        /// <summary>
        /// 진행 바 브러시
        /// </summary>
        [Category("WaveformTimeline")]
        public Brush ProgressBarBrush
        {
            get
            {
                return (Brush)GetValue(ProgressBarBrushProperty);
            }
            set
            {
                SetValue(ProgressBarBrushProperty, value);
            }
        }

        #endregion
        #region 진행 바 두께 - ProgressBarThickness

        /// <summary>
        /// 진행 바 두께
        /// </summary>
        [Category("WaveformTimeline")]
        public double ProgressBarThickness
        {
            get
            {
                return (double)GetValue(ProgressBarThicknessProperty);
            }
            set
            {
                SetValue(ProgressBarThicknessProperty, value);
            }
        }

        #endregion
        #region 중앙 라인 브러시 - CenterLineBrush

        /// <summary>
        /// 중앙 라인 브러시
        /// </summary>
        [Category("WaveformTimeline")]
        public Brush CenterLineBrush
        {
            get
            {
                return (Brush)GetValue(CenterLineBrushProperty);
            }
            set
            {
                SetValue(CenterLineBrushProperty, value);
            }
        }

        #endregion
        #region 중앙 라인 두께 - CenterLineThickness

        /// <summary>
        /// 중앙 라인 두께
        /// </summary>
        [Category("WaveformTimeline")]
        public double CenterLineThickness
        {
            get
            {
                return (double)GetValue(CenterLineThicknessProperty);
            }
            set
            {
                SetValue(CenterLineThicknessProperty, value);
            }
        }

        #endregion
        #region 반복 영역 브러시 - RepeatRegionBrush

        /// <summary>
        /// 반복 영역 브러시
        /// </summary>
        [Category("WaveformTimeline")]
        public Brush RepeatRegionBrush
        {
            get
            {
                return (Brush)GetValue(RepeatRegionBrushProperty);
            }
            set
            {
                SetValue(RepeatRegionBrushProperty, value);
            }
        }

        #endregion
        #region 반복 영역 허용 여부 - AllowRepeatRegion

        /// <summary>
        /// 반복 영역 허용 여부
        /// </summary>
        [Category("WaveformTimeline")]
        public bool AllowRepeatRegion
        {
            get
            {
                return (bool)GetValue(AllowRepeatRegionProperty);
            }
            set
            {
                SetValue(AllowRepeatRegionProperty, value);
            }
        }

        #endregion
        #region 타임라인 틱 브러시 - TimelineTickBrush

        /// <summary>
        /// 타임라인 틱 브러시
        /// </summary>
        [Category("WaveformTimeline")]
        public Brush TimelineTickBrush
        {
            get
            {
                return (Brush)GetValue(TimelineTickBrushProperty);
            }
            set
            {
                SetValue(TimelineTickBrushProperty, value);
            }
        }

        #endregion
        #region 오토 스케일 웨이브폼 캐시 - AutoScaleWaveformCache

        /// <summary>
        /// 오토 스케일 웨이브폼 캐시
        /// </summary>
        [Category("Common")]
        public bool AutoScaleWaveformCache
        {
            get
            {
                return (bool)GetValue(AutoScaleWaveformCacheProperty);
            }
            set
            {
                SetValue(AutoScaleWaveformCacheProperty, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Static

        #region 생성자 - WaveformTimeline()

        /// <summary>
        /// 생성자
        /// </summary>
        static WaveformTimeline()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(WaveformTimeline), new FrameworkPropertyMetadata(typeof(WaveformTimeline)));
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 왼쪽 레벨 브러시 속성 변경시 콜백 처리하기 - LeftLevelBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 왼쪽 레벨 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void LeftLevelBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnLeftLevelBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 왼쪽 레벨 브러시 속성 강제시 콜백 처리하기 - CoerceLeftLevelBrushPropertyCallback(d, value)

        /// <summary>
        /// 왼쪽 레벨 브러시 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>수정 값</returns>
        private static object CoerceLeftLevelBrushPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceLeftLevelBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 오른쪽 레벨 브러시 속성 변경시 콜백 처리하기 - RightLevelBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 오른쪽 레벨 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void RightLevelBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnRightLevelBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 오른쪽 레벨 브러시 속성 강제시 콜백 처리하기 - CoerceRightLevelBrushPropertyCallback(d, value)

        /// <summary>
        /// 오른쪽 레벨 브러시 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceRightLevelBrushPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceRightLevelBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 진행 바 브러시 속성 변경시 콜백 처리하기 - ProgressBarBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 진행 바 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ProgressBarBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnProgressBarBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 진행 바 브러시 강제시 콜백 처리하기 - CoerceProgressBarBrushPropertyCallback(d, value)

        /// <summary>
        /// 진행 바 브러시 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceProgressBarBrushPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceProgressBarBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 진행 바 두께 속성 변경시 콜백 처리하기 - ProgressBarThicknessPropertyChangedCallback(d, e)

        /// <summary>
        /// 진행 바 두께 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ProgressBarThicknessPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnProgressBarThicknessChanged((double)e.OldValue, (double)e.NewValue);
            }
        }

        #endregion
        #region 진행 바 두께 속성 강제시 콜백 처리하기 - CoerceProgressBarThicknessPropertyCallback(d, value)

        /// <summary>
        /// 진행 바 두께 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceProgressBarThicknessPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceProgressBarThickness((double)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 중앙 라인 브러시 속성 변경시 콜백 처리하기 - CenterLineBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 중앙 라인 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void CenterLineBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnCenterLineBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 중앙 라인 브러시 속성 강제시 콜백 처리하기 - CoerceCenterLineBrushPropertyCallback(d, value)

        /// <summary>
        /// 중앙 라인 브러시 속성 강제시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceCenterLineBrushPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceCenterLineBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 중앙 라인 두께 속성 변경시 콜백 처리하기 - CenterLineThicknessPropertyChangedCallback(d, e)

        /// <summary>
        /// 중앙 라인 두께 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void CenterLineThicknessPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnCenterLineThicknessChanged((double)e.OldValue, (double)e.NewValue);
            }
        }

        #endregion
        #region 중앙 라인 두께 속성 강제시 콜배 처리하기 - CoerceCenterLineThicknessPropertyCallback(d, value)

        /// <summary>
        /// 중앙 라인 두께 속성 강제시 콜배 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceCenterLineThicknessPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceCenterLineThickness((double)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 반복 영역 브러시 속성 변경시 콜백 처리하기 - RepeatRegionBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 반복 영역 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void RepeatRegionBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnRepeatRegionBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 반복 영역 브러시 속성 강제시 콜배 처리하기 - CoerceRepeatRegionBrushPropertyCallback(d, value)

        /// <summary>
        /// 반복 영역 브러시 속성 강제시 콜배 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceRepeatRegionBrushPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceRepeatRegionBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 반복 영역 허용 여부 속성 변경시 콜백 처리하기 - AllowRepeatRegionPropertyChangedCallback(d, e)

        /// <summary>
        /// 반복 영역 허용 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void AllowRepeatRegionPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnAllowRepeatRegionChanged((bool)e.OldValue, (bool)e.NewValue);
            }
        }

        #endregion
        #region 반복 영역 허용 여부 속성 강제시 콜배 처리하기 - CoerceAllowRepeatRegionPropertyCallback(d, value)

        /// <summary>
        /// 반복 영역 허용 여부 속성 강제시 콜배 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceAllowRepeatRegionPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceAllowRepeatRegion((bool)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 타임라인 틱 브러시 속성 변경시 콜백 처리하기 - TimelineTickBrushPropertyChangedCallback(d, e)

        /// <summary>
        /// 타임라인 틱 브러시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void TimelineTickBrushPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnTimelineTickBrushChanged((Brush)e.OldValue, (Brush)e.NewValue);
            }
        }

        #endregion
        #region 타임라인 틱 브러시 속성 강제시 콜배 처리하기 - CoerceTimelineTickBrushPropertyCallback(d, value)

        /// <summary>
        /// 타임라인 틱 브러시 속성 강제시 콜배 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceTimelineTickBrushPropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceTimelineTickBrush((Brush)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        #region 오토 스케일 웨이브폼 캐시 속성 변경시 콜백 처리하기 - AutoScaleWaveformCachePropertyChangedCallback(d, e)

        /// <summary>
        /// 오토 스케일 웨이브폼 캐시 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void AutoScaleWaveformCachePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                waveformTimeline.OnAutoScaleWaveformCacheChanged((bool)e.OldValue, (bool)e.NewValue);
            }
        }

        #endregion
        #region 오토 스케일 웨이브폼 캐시 속성 강제시 콜배 처리하기 - CoerceAutoScaleWaveformCachePropertyCallback(d, value)

        /// <summary>
        /// 오토 스케일 웨이브폼 캐시 속성 강제시 콜배 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        private static object CoerceAutoScaleWaveformCachePropertyCallback(DependencyObject d, object value)
        {
            WaveformTimeline waveformTimeline = d as WaveformTimeline;

            if(waveformTimeline != null)
            {
                return waveformTimeline.OnCoerceAutoScaleWaveformCache((bool)value);
            }
            else
            {
                return value;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 템플리트 적용시 처리하기 - OnApplyTemplate()

        /// <summary>
        /// 템플리트 적용시 처리하기
        /// </summary>
        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            this.waveformCanvas = GetTemplateChild("PART_Waveform") as Canvas;

            this.waveformCanvas.CacheMode = new BitmapCache();

            this.waveformCanvas.Background = new SolidColorBrush(Colors.Transparent);

            this.waveformCanvas.Children.Add(this.centerLine);
            this.waveformCanvas.Children.Add(this.leftPath  );
            this.waveformCanvas.Children.Add(this.rightPath );

            this.timelineCanvas = GetTemplateChild("PART_Timeline") as Canvas;

            this.timelineCanvas.Children.Add(this.timelineBackgroundRegion);

            this.timelineCanvas.SizeChanged += timelineCanvas_SizeChanged;

            this.repeatCanvas = GetTemplateChild("PART_Repeat") as Canvas;

            this.repeatCanvas.Children.Add(this.repeatRegion);

            this.progressCanvas = GetTemplateChild("PART_Progress") as Canvas;

            this.progressCanvas.Children.Add(this.progressIndicator);
            this.progressCanvas.Children.Add(this.progressLine     );

            UpdateWaveformCacheScaling();            
        }

        #endregion

        #region 사운드 플레이어 등록하기 - RegisterSoundPlayer(soundPlayer)

        /// <summary>
        /// 사운드 플레이어 등록하기
        /// </summary>
        /// <param name="soundPlayer">사운드 플레이어 인터페이스</param>
        public void RegisterSoundPlayer(IWaveformPlayer soundPlayer)
        {
            this.soundPlayer = soundPlayer;

            soundPlayer.PropertyChanged += soundPlayer_PropertyChanged;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Protected

        #region 왼쪽 레벨 브러시 변경시 처리하기 - OnLeftLevelBrushChanged(oldValue, newValue)

        /// <summary>
        /// 왼쪽 레벨 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnLeftLevelBrushChanged(Brush oldValue, Brush newValue)
        {
            this.leftPath.Fill = LeftLevelBrush;

            UpdateWaveform();
        }

        #endregion
        #region 왼쪽 레벨 브러시 강제시 처리하기 - OnCoerceLeftLevelBrush(value)

        /// <summary>
        /// 왼쪽 레벨 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceLeftLevelBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 오른쪽 레벨 브러시 변경시 처리하기 - OnRightLevelBrushChanged(oldValue, newValue)

        /// <summary>
        /// 오른쪽 레벨 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnRightLevelBrushChanged(Brush oldValue, Brush newValue)
        {
            this.rightPath.Fill = RightLevelBrush;

            UpdateWaveform();
        }

        #endregion
        #region 오른쪽 레벨 브러시 강제시 처리하기 - OnCoerceRightLevelBrush(value)

        /// <summary>
        /// 오른쪽 레벨 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceRightLevelBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 진행 바 브러시 변경시 처리하기 - OnProgressBarBrushChanged(oldValue, newValue)

        /// <summary>
        /// 진행 바 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnProgressBarBrushChanged(Brush oldValue, Brush newValue)
        {
            this.progressIndicator.Fill = ProgressBarBrush;

            this.progressLine.Stroke = ProgressBarBrush;

            CreateProgressIndicator();
        }

        #endregion
        #region 진행 바 브러시 강제시 처리하기 - OnCoerceProgressBarBrush(value)

        /// <summary>
        /// 진행 바 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceProgressBarBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 진행 바 두께 변경시 처리하기 - OnProgressBarThicknessChanged(oldValue, newValue)

        /// <summary>
        /// 진행 바 두께 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnProgressBarThicknessChanged(double oldValue, double newValue)
        {
            this.progressLine.StrokeThickness = ProgressBarThickness;

            CreateProgressIndicator();
        }

        #endregion
        #region 진행 바 두께 강제시 처리하기 - OnCoerceProgressBarThickness(value)

        /// <summary>
        /// 진행 바 두께 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual double OnCoerceProgressBarThickness(double value)
        {
            value = Math.Max(value, 0.0d);

            return value;
        }

        #endregion

        #region 중앙 라인 브러시 변경시 처리하기 - OnCenterLineBrushChanged(oldValue, newValue)

        /// <summary>
        /// 중앙 라인 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnCenterLineBrushChanged(Brush oldValue, Brush newValue)
        {
            this.centerLine.Stroke = CenterLineBrush;

            UpdateWaveform();
        }

        #endregion
        #region 중앙 라인 브러시 강제시 처리하기 - OnCoerceCenterLineBrush(value)

        /// <summary>
        /// 중앙 라인 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceCenterLineBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 중앙 라인 두께 변경시 처리하기 - OnCenterLineThicknessChanged(oldValue, newValue)

        /// <summary>
        /// 중앙 라인 두께 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnCenterLineThicknessChanged(double oldValue, double newValue)
        {
            this.centerLine.StrokeThickness = CenterLineThickness;

            UpdateWaveform();
        }

        #endregion
        #region 중앙 라인 두께 강제시 처리하기 - OnCoerceCenterLineThickness(value)

        /// <summary>
        /// 중앙 라인 두께 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual double OnCoerceCenterLineThickness(double value)
        {
            value = Math.Max(value, 0.0d);

            return value;
        }

        #endregion

        #region 반복 영역 브러시 변경시 처리하기 - OnRepeatRegionBrushChanged(oldValue, newValue)

        /// <summary>
        /// 반복 영역 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnRepeatRegionBrushChanged(Brush oldValue, Brush newValue)
        {
            this.repeatRegion.Fill = RepeatRegionBrush;

            UpdateRepeatRegion();
        }

        #endregion
        #region 반복 영역 브러시 강제시 처리하기 - OnCoerceRepeatRegionBrush(value)

        /// <summary>
        /// 반복 영역 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceRepeatRegionBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 반복 영역 허용 여부 변경시 처리하기 - OnAllowRepeatRegionChanged(oldValue, newValue)

        /// <summary>
        /// 반복 영역 허용 여부 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnAllowRepeatRegionChanged(bool oldValue, bool newValue)
        {
            if(!newValue && this.soundPlayer != null)
            {
                this.soundPlayer.SelectionBegin = TimeSpan.Zero;
                this.soundPlayer.SelectionEnd   = TimeSpan.Zero;
            }
        }

        #endregion
        #region 반복 영역 허용 여부 강제시 처리하기 - OnCoerceAllowRepeatRegion(value)

        /// <summary>
        /// 반복 영역 허용 여부 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual bool OnCoerceAllowRepeatRegion(bool value)
        {
            return value;
        }

        #endregion

        #region 타임라인 틱 브러시 변경시 처리하기 - OnTimelineTickBrushChanged(oldValue, newValue)

        /// <summary>
        /// 타임라인 틱 브러시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnTimelineTickBrushChanged(Brush oldValue, Brush newValue)
        {
            UpdateTimeline();
        }

        #endregion
        #region 타임라인 틱 브러시 강제시 처리하기 - OnCoerceTimelineTickBrush(value)

        /// <summary>
        /// 타임라인 틱 브러시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual Brush OnCoerceTimelineTickBrush(Brush value)
        {
            return value;
        }

        #endregion

        #region 오토 스케일 웨이브폼 캐시 변경시 처리하기 - OnAutoScaleWaveformCacheChanged(oldValue, newValue)

        /// <summary>
        /// 오토 스케일 웨이브폼 캐시 변경시 처리하기
        /// </summary>
        /// <param name="oldValue">이전 값</param>
        /// <param name="newValue">신규 값</param>
        protected virtual void OnAutoScaleWaveformCacheChanged(bool oldValue, bool newValue)
        {
            UpdateWaveformCacheScaling();
        }

        #endregion
        #region 오토 스케일 웨이브폼 캐시 강제시 처리하기 - OnCoerceAutoScaleWaveformCache(value)

        /// <summary>
        /// 오토 스케일 웨이브폼 캐시 강제시 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>강제 값</returns>
        protected virtual bool OnCoerceAutoScaleWaveformCache(bool value)
        {
            return value;
        }

        #endregion

        #region 템플리트 변경시 처리하기 - OnTemplateChanged(oldTemplate, newTemplate)

        /// <summary>
        /// 템플리트 변경시 처리하기
        /// </summary>
        /// <param name="oldTemplate">이전 템플리트</param>
        /// <param name="newTemplate">신규 템플리트</param>
        protected override void OnTemplateChanged(ControlTemplate oldTemplate, ControlTemplate newTemplate)
        {
            base.OnTemplateChanged(oldTemplate, newTemplate);

            if(this.waveformCanvas != null)
            {
                this.waveformCanvas.Children.Clear();
            }

            if(this.timelineCanvas != null)
            {
                this.timelineCanvas.SizeChanged -= timelineCanvas_SizeChanged;

                this.timelineCanvas.Children.Clear();
            }

            if(this.repeatCanvas != null)
            {
                this.repeatCanvas.Children.Clear();
            }

            if(this.progressCanvas != null)
            {
                this.progressCanvas.Children.Clear();
            }
        }

        #endregion
        #region 렌더 크기 변경시 처리하기 - OnRenderSizeChanged(info)

        /// <summary>
        /// 렌더 크기 변경시 처리하기
        /// </summary>
        /// <param name="info">크기 변경시 정보</param>
        protected override void OnRenderSizeChanged(SizeChangedInfo info)
        {
            base.OnRenderSizeChanged(info);

            UpdateWaveformCacheScaling();

            UpdateAllRegions();
        }

        #endregion

        #region 마우스 왼쪽 버튼 DOWN 처리하기 - OnMouseLeftButtonDown(e)

        /// <summary>
        /// 마우스 왼쪽 버튼 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonDown(e);

            CaptureMouse();

            this.isMouseDown = true;

            this.mouseDownPoint = e.GetPosition(this.waveformCanvas);
        }

        #endregion
        #region 마우스 왼쪽 버튼 UP 처리하기 - OnMouseLeftButtonUp(e)

        /// <summary>
        /// 마우스 왼쪽 버튼 UP 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonUp(e);            

            if(!this.isMouseDown)
            {
                return;
            }

            bool updateRepeatRegion = false;

            this.isMouseDown = false;

            ReleaseMouseCapture();

            if(Math.Abs(this.currentPoint.X - this.mouseDownPoint.X) < MOUSE_MOVE_TOLERANCE)
            {
                if(IsPointInRepeatRegion(this.mouseDownPoint))
                {
                    double position = (this.currentPoint.X / RenderSize.Width) * this.soundPlayer.ChannelLength;

                    this.soundPlayer.ChannelPosition = Math.Min(this.soundPlayer.ChannelLength, Math.Max(0, position));
                }
                else
                {
                    this.soundPlayer.SelectionBegin = TimeSpan.Zero;
                    this.soundPlayer.SelectionEnd   = TimeSpan.Zero;

                    double position = (this.currentPoint.X / RenderSize.Width) * this.soundPlayer.ChannelLength;

                    this.soundPlayer.ChannelPosition = Math.Min(this.soundPlayer.ChannelLength, Math.Max(0, position));

                    this.startLoopRegion = -1;
                    this.endLoopRegion   = -1;

                    updateRepeatRegion = true;
                }
            }
            else
            {
                this.soundPlayer.SelectionBegin = TimeSpan.FromSeconds(this.startLoopRegion);
                this.soundPlayer.SelectionEnd   = TimeSpan.FromSeconds(this.endLoopRegion  );

                double position = this.startLoopRegion;

                this.soundPlayer.ChannelPosition = Math.Min(this.soundPlayer.ChannelLength, Math.Max(0, position));

                updateRepeatRegion = true;
            }

            if(updateRepeatRegion)
            {
                UpdateRepeatRegion();
            }
        }

        #endregion
        #region 마우스 이동시 처리하기 - OnMouseMove(e)

        /// <summary>
        /// 마우스 이동시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            this.currentPoint = e.GetPosition(this.waveformCanvas);

            if(this.isMouseDown && AllowRepeatRegion)
            {
                if(Math.Abs(this.currentPoint.X - this.mouseDownPoint.X) > MOUSE_MOVE_TOLERANCE)
                {
                    if(this.mouseDownPoint.X < this.currentPoint.X)
                    {
                        this.startLoopRegion = (this.mouseDownPoint.X / RenderSize.Width) * this.soundPlayer.ChannelLength;
                        this.endLoopRegion   = (this.currentPoint.X   / RenderSize.Width) * this.soundPlayer.ChannelLength;
                    }
                    else
                    {
                        this.startLoopRegion = (this.currentPoint.X   / RenderSize.Width) * this.soundPlayer.ChannelLength;
                        this.endLoopRegion   = (this.mouseDownPoint.X / RenderSize.Width) * this.soundPlayer.ChannelLength;
                    }
                }
                else
                {
                    this.startLoopRegion = -1;
                    this.endLoopRegion   = -1;
                }

                UpdateRepeatRegion();
            }
        }
        
        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Event

        #region 타임라인 캔버스 크기 변경시 처리하기 - timelineCanvas_SizeChanged(sender, e)

        /// <summary>
        /// 타임라인 캔버스 크기 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timelineCanvas_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            UpdateTimeline();
        }

        #endregion
        #region 사운드 플레이어 속성 변경시 처리하기 - soundPlayer_PropertyChanged(sender, e)

        /// <summary>
        /// 사운드 플레이어 속성 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void soundPlayer_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            switch(e.PropertyName)
            {
                case "SelectionBegin" :

                    this.startLoopRegion = this.soundPlayer.SelectionBegin.TotalSeconds;

                    UpdateRepeatRegion();

                    break;

                case "SelectionEnd" :

                    this.endLoopRegion = this.soundPlayer.SelectionEnd.TotalSeconds;

                    UpdateRepeatRegion();

                    break;

                case "WaveformData" :

                    UpdateWaveform();

                    break;

                case "ChannelPosition" :

                    UpdateProgressIndicator();

                    break;

                case "ChannelLength" :

                    this.startLoopRegion = -1;
                    this.endLoopRegion   = -1;

                    UpdateAllRegions();

                    break;
            }
        }

        #endregion

        #region 반복 영역의 포인트 여부 구하기 - IsPointInRepeatRegion(point)

        /// <summary>
        /// 반복 영역의 포인트 여부 구하기
        /// </summary>
        /// <param name="point">포인트</param>
        /// <returns>반복 영역의 포인트 여부</returns>
        private bool IsPointInRepeatRegion(Point point)
        {
            if(this.soundPlayer.ChannelLength == 0)
            {
                return false;
            }

            double regionLeft  = (this.soundPlayer.SelectionBegin.TotalSeconds / this.soundPlayer.ChannelLength) * RenderSize.Width;
            double regionRight = (this.soundPlayer.SelectionEnd.TotalSeconds / this.soundPlayer.ChannelLength) * RenderSize.Width;

            return (point.X >= regionLeft && point.X < regionRight);
        }

        #endregion

        #region 전체 변환 스케일 구하기 - GetTotalTransformScale()

        /// <summary>
        /// 전체 변환 스케일 구하기
        /// </summary>
        /// <returns>전체 변환 스케일</returns>
        private double GetTotalTransformScale()
        {
            double totalTransform = 1.0d;

            DependencyObject dependencyObject = this;

            do
            {
                Visual visual = dependencyObject as Visual;

                if(visual != null)
                {
                    Transform transform = VisualTreeHelper.GetTransform(visual);

                    if((transform != null) && (transform.Value.M12 == 0) && (transform.Value.M21 == 0) && (transform.Value.OffsetX == 0) && (transform.Value.OffsetY == 0) && (transform.Value.M11 == transform.Value.M22))
                    {
                        totalTransform *= transform.Value.M11;
                    }
                }

                dependencyObject = VisualTreeHelper.GetParent(dependencyObject);
            }
            while(dependencyObject != null);

            return totalTransform;
        }

        #endregion
        #region 웨이브폼 캐시 스케일 업데이트하기 - UpdateWaveformCacheScaling()

        /// <summary>
        /// 웨이브폼 캐시 스케일 업데이트하기
        /// </summary>
        private void UpdateWaveformCacheScaling()
        {
            if(this.waveformCanvas == null)
            {
                return;
            }

            BitmapCache bitmapCache = (BitmapCache)this.waveformCanvas.CacheMode;

            if(AutoScaleWaveformCache)
            {
                double totalTransformScale = GetTotalTransformScale();

                if(bitmapCache.RenderAtScale != totalTransformScale)
                {
                    bitmapCache.RenderAtScale = totalTransformScale;
                }
            }
            else
            {
                bitmapCache.RenderAtScale = 1.0d;
            }
        }

        #endregion

        #region 반복 영역 업데이트하기 - UpdateRepeatRegion()

        /// <summary>
        /// 반복 영역 업데이트하기
        /// </summary>
        private void UpdateRepeatRegion()
        {
            if(this.soundPlayer == null || this.repeatCanvas == null)
            {
                return;
            }

            double startPercent = this.startLoopRegion / this.soundPlayer.ChannelLength;
            double startX       = startPercent * this.repeatCanvas.RenderSize.Width;
            double endPercent   = this.endLoopRegion / this.soundPlayer.ChannelLength;
            double endX         = endPercent * this.repeatCanvas.RenderSize.Width;

            if(this.soundPlayer.ChannelLength == 0 || endX <= startX)
            {
                this.repeatRegion.Width  = 0;
                this.repeatRegion.Height = 0;

                return;
            }
            
            this.repeatRegion.Margin = new Thickness(startX, 0, 0, 0);
            this.repeatRegion.Width  = endX - startX;
            this.repeatRegion.Height = this.repeatCanvas.RenderSize.Height;
        }

        #endregion
        #region 진행 지시자 업데이트하기 - UpdateProgressIndicator()

        /// <summary>
        /// 진행 지시자 업데이트하기
        /// </summary>
        private void UpdateProgressIndicator()
        {
            if(this.soundPlayer == null || this.progressCanvas == null)
            {
                return;
            }

            double x = 0.0d;

            if(this.soundPlayer.ChannelLength != 0)
            {
                double progressPercent = this.soundPlayer.ChannelPosition / this.soundPlayer.ChannelLength;

                x = progressPercent * this.progressCanvas.RenderSize.Width;
            }

            this.progressLine.Margin      = new Thickness(x, 0, 0, 0);
            this.progressIndicator.Margin = new Thickness(x, 0, 0, 0);
        }

        #endregion
        #region 진행 지시자 생성하기 - CreateProgressIndicator()

        /// <summary>
        /// 진행 지시자 생성하기
        /// </summary>
        private void CreateProgressIndicator()
        {
            if(this.soundPlayer == null || this.timelineCanvas == null || this.progressCanvas == null)
            {
                return;
            }

            const double x = 0.0d;

            this.progressLine.X1 = x;
            this.progressLine.X2 = x;
            this.progressLine.Y1 = this.timelineCanvas.RenderSize.Height;
            this.progressLine.Y2 = this.progressCanvas.RenderSize.Height;

            PolyLineSegment indicatorPolyLineSegment = new PolyLineSegment();

            indicatorPolyLineSegment.Points.Add(new Point(x                           , this.timelineCanvas.RenderSize.Height                           ));
            indicatorPolyLineSegment.Points.Add(new Point(x - INDICATOR_TRIANGLE_WIDTH, this.timelineCanvas.RenderSize.Height - INDICATOR_TRIANGLE_WIDTH));
            indicatorPolyLineSegment.Points.Add(new Point(x + INDICATOR_TRIANGLE_WIDTH, this.timelineCanvas.RenderSize.Height - INDICATOR_TRIANGLE_WIDTH));
            indicatorPolyLineSegment.Points.Add(new Point(x                           , this.timelineCanvas.RenderSize.Height                           ));

            PathGeometry indicatorPathGeometry = new PathGeometry();

            PathFigure indicatorPathFigure = new PathFigure();

            indicatorPathFigure.Segments.Add(indicatorPolyLineSegment);

            indicatorPathGeometry.Figures.Add(indicatorPathFigure);

            this.progressIndicator.Data = indicatorPathGeometry;

            UpdateProgressIndicator();
        }

        #endregion
        #region 타임라인 업데이트하기 - UpdateTimeline()

        /// <summary>
        /// 타임라인 업데이트하기
        /// </summary>
        private void UpdateTimeline()
        {
            if(this.soundPlayer == null || this.timelineCanvas == null)
            {
                return;
            }

            foreach(TextBlock textblock in this.timestampTextBlockList)
            {
                this.timelineCanvas.Children.Remove(textblock);
            }

            this.timestampTextBlockList.Clear();

            foreach(Line line in this.timeLineTickList)
            {
                this.timelineCanvas.Children.Remove(line);
            }

            this.timeLineTickList.Clear();

            double bottom = this.timelineCanvas.RenderSize.Height - 1;

            this.timelineBackgroundRegion.Width  = this.timelineCanvas.RenderSize.Width;
            this.timelineBackgroundRegion.Height = this.timelineCanvas.RenderSize.Height;

            double minorTickDuration = 1.00d;
            double majorTickDuration = 5.00d;

            if(this.soundPlayer.ChannelLength >= 120.0d)
            {
                minorTickDuration = 15.0d;
                majorTickDuration = 60.0d;
            }
            else if(this.soundPlayer.ChannelLength >= 60.0d)
            {
                minorTickDuration = 5.0d;
                majorTickDuration = 30.0d;
            }
            else if(this.soundPlayer.ChannelLength >= 30.0d)
            {
                minorTickDuration = 2.0d;
                majorTickDuration = 10.0d;
            }

            if(this.soundPlayer.ChannelLength < minorTickDuration)
            {
                return;
            }

            int minorTickCount = (int)(this.soundPlayer.ChannelLength / minorTickDuration);

            for(int i = 1; i <= minorTickCount; i++)
            {
                Line timelineTick = new Line()
                {
                    Stroke          = TimelineTickBrush,
                    StrokeThickness = 1.0d
                };

                if(i % (majorTickDuration / minorTickDuration) == 0)
                {
                    double x = ((i * minorTickDuration) / this.soundPlayer.ChannelLength) * this.timelineCanvas.RenderSize.Width;

                    bool drawTextBlock = false;

                    double lastEndTimestamp;

                    if(this.timestampTextBlockList.Count != 0)
                    {
                        TextBlock lastTextBlock = this.timestampTextBlockList[this.timestampTextBlockList.Count - 1];

                        lastEndTimestamp = lastTextBlock.Margin.Left + lastTextBlock.ActualWidth;
                    }
                    else
                    {
                        lastEndTimestamp = 0;
                    }

                    if(x > lastEndTimestamp + TIMESTAMP_MARGIN)
                    {
                        drawTextBlock = true;
                    }

                    bool isEndOfTimeline = (this.timelineCanvas.RenderSize.Width - x < 28.0d);

                    if(drawTextBlock)
                    {
                        timelineTick.X1 = x;
                        timelineTick.Y1 = bottom;
                        timelineTick.X2 = x;
                        timelineTick.Y2 = bottom - MAJOR_TICK_HEIGHT;

                        if(isEndOfTimeline)
                        {
                            continue;
                        }

                        TimeSpan timeSpan = TimeSpan.FromSeconds(i * minorTickDuration);

                        TextBlock timestampText = new TextBlock()
                        {
                            Margin      = new Thickness(x + 2, 0, 0, 0),
                            FontFamily  = this.FontFamily,
                            FontStyle   = this.FontStyle,
                            FontWeight  = this.FontWeight,
                            FontStretch = this.FontStretch,
                            FontSize    = this.FontSize,
                            Foreground  = this.Foreground,
                            Text        = (timeSpan.TotalHours >= 1.0d) ? string.Format(@"{0:hh\:mm\:ss}", timeSpan) : string.Format(@"{0:mm\:ss}", timeSpan)
                        };

                        this.timestampTextBlockList.Add(timestampText);

                        this.timelineCanvas.Children.Add(timestampText);

                        UpdateLayout();
                    }
                    else
                    {
                        timelineTick.X1 = x;
                        timelineTick.Y1 = bottom;
                        timelineTick.X2 = x;
                        timelineTick.Y2 = bottom - MINOR_TICK_HEIGHT;
                    }
                }
                else
                {
                    double x = ((i * minorTickDuration) / this.soundPlayer.ChannelLength) * this.timelineCanvas.RenderSize.Width;

                    timelineTick.X1 = x;
                    timelineTick.Y1 = bottom;
                    timelineTick.X2 = x;
                    timelineTick.Y2 = bottom - MINOR_TICK_HEIGHT;
                }

                this.timeLineTickList.Add(timelineTick);

                this.timelineCanvas.Children.Add(timelineTick);
            }
        }

        #endregion
        #region 웨이브폼 업데이트하기 - UpdateWaveform()

        /// <summary>
        /// 웨이브폼 업데이트하기
        /// </summary>
        private void UpdateWaveform()
        {
            const double minumValue   = 0;
            const double maximumValue = 1.5;
            const double scale        = (maximumValue - minumValue);

            if(this.soundPlayer == null || this.soundPlayer.WaveFormFloatArray == null || this.waveformCanvas == null || this.waveformCanvas.RenderSize.Width < 1 || this.waveformCanvas.RenderSize.Height < 1)
            {
                return;
            }

            double leftRenderHeight;
            double rightRenderHeight;

            int pointCount = (int)(this.soundPlayer.WaveFormFloatArray.Length / 2.0d);

            double pointThickness     = this.waveformCanvas.RenderSize.Width / pointCount;
            double waveformSideHeight = this.waveformCanvas.RenderSize.Height / 2.0d;
            double centerHeight       = waveformSideHeight;

            if(CenterLineBrush != null)
            {
                this.centerLine.X1 = 0;
                this.centerLine.X2 = this.waveformCanvas.RenderSize.Width;
                this.centerLine.Y1 = centerHeight;
                this.centerLine.Y2 = centerHeight;
            }

            if(this.soundPlayer.WaveFormFloatArray != null && this.soundPlayer.WaveFormFloatArray.Length > 1)
            {
                PolyLineSegment leftWaveformPolyLine = new PolyLineSegment();

                leftWaveformPolyLine.Points.Add(new Point(0, centerHeight));

                PolyLineSegment rightWaveformPolyLine = new PolyLineSegment();

                rightWaveformPolyLine.Points.Add(new Point(0, centerHeight));

                double x = 0.0d;

                for(int i = 0; i < this.soundPlayer.WaveFormFloatArray.Length; i += 2)
                {
                    x = (i / 2) * pointThickness;

                    leftRenderHeight = ((this.soundPlayer.WaveFormFloatArray[i] - minumValue) / scale) * waveformSideHeight;

                    leftWaveformPolyLine.Points.Add(new Point(x, centerHeight - leftRenderHeight));

                    rightRenderHeight = ((this.soundPlayer.WaveFormFloatArray[i + 1] - minumValue) / scale) * waveformSideHeight;

                    rightWaveformPolyLine.Points.Add(new Point(x, centerHeight + rightRenderHeight));
                }

                leftWaveformPolyLine.Points.Add(new Point(x, centerHeight));
                leftWaveformPolyLine.Points.Add(new Point(0, centerHeight));

                rightWaveformPolyLine.Points.Add(new Point(x, centerHeight));
                rightWaveformPolyLine.Points.Add(new Point(0, centerHeight));

                PathGeometry leftGeometry = new PathGeometry();

                PathFigure leftPathFigure = new PathFigure();

                leftPathFigure.Segments.Add(leftWaveformPolyLine);

                leftGeometry.Figures.Add(leftPathFigure);

                PathGeometry rightGeometry = new PathGeometry();

                PathFigure rightPathFigure = new PathFigure();

                rightPathFigure.Segments.Add(rightWaveformPolyLine);

                rightGeometry.Figures.Add(rightPathFigure);

                this.leftPath.Data  = leftGeometry;
                this.rightPath.Data = rightGeometry;
            }
            else
            {
                this.leftPath.Data  = null;
                this.rightPath.Data = null;
            }
        }

        #endregion
        #region 모든 영역 업데이트하기 - UpdateAllRegions()

        /// <summary>
        /// 모든 영역 업데이트하기
        /// </summary>
        private void UpdateAllRegions()
        {
            UpdateRepeatRegion();

            CreateProgressIndicator();

            UpdateTimeline();

            UpdateWaveform();
        }

        #endregion
    }
}