﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Windows.Threading;

using NAudio.Wave;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    /// NAUDIO 엔진
    /// </summary>
    public class NAudioEngine : INotifyPropertyChanged, ISpectrumPlayer, IWaveformPlayer, IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 웨이브폼 생성 매개 변수 - WaveformGenerationParameter

        /// <summary>
        /// 웨이브폼 생성 매개 변수
        /// </summary>
        private class WaveformGenerationParameter
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 포인트 수 - PointCount

            /// <summary>
            /// 포인트 수
            /// </summary>
            public int PointCount { get; protected set; }

            #endregion
            #region 경로 - Path

            /// <summary>
            /// 경로
            /// </summary>
            public string Path { get; protected set; }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - WaveformGenerationParameter(pointCount, path)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="pointCount">포인트 수</param>
            /// <param name="path">경로</param>
            public WaveformGenerationParameter(int pointCount, string path)
            {
                PointCount = pointCount;
                Path       = path;
            }

            #endregion

        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (INotifyPropertyChanged) 속성 변경시 - PropertyChanged

        /// <summary>
        /// 속성 변경시
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 인스턴스
        /// </summary>
        private static NAudioEngine _instance;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 웨이브폼 압축 포인트 수
        /// </summary>
        private const int WAVEFORM_COMPRESSED_POINT_COUNT = 2000;

        /// <summary>
        /// 반복 한계점
        /// </summary>
        private const int REPEAT_THRESHOLD = 200;


        /// <summary>
        /// 위치 디스패처 타이머
        /// </summary>
        private readonly DispatcherTimer positionDispatcherTimer = new DispatcherTimer(DispatcherPriority.ApplicationIdle);

        /// <summary>
        /// 웨이프 폼 생성 백그라운드 작업자
        /// </summary>
        private readonly BackgroundWorker waveformGenerateBackgroundWorker = new BackgroundWorker();

        /// <summary>
        /// FFT 데이터 크기
        /// </summary>
        private readonly int fftDataSize = (int)FFTDataSize.FFT2048;

        /// <summary>
        /// 리소스 해제 여부
        /// </summary>
        private bool disposed;

        /// <summary>
        /// 재생 가능 여부
        /// </summary>
        private bool canPlay;

        /// <summary>
        /// 중지 가능 여부
        /// </summary>
        private bool canPause;

        /// <summary>
        /// 중단 가능 여부
        /// </summary>
        private bool canStop;

        /// <summary>
        /// 재생 여부
        /// </summary>
        private bool isPlaying;

        /// <summary>
        /// 채널 타이머 업데이트 중 여부
        /// </summary>
        private bool inChannelTimerUpdate;

        /// <summary>
        /// 채널 길이
        /// </summary>
        private double channelLength;

        /// <summary>
        /// 채널 위치
        /// </summary>
        private double channelPosition;

        /// <summary>
        /// 채널 설정 중 여부
        /// </summary>
        private bool inChannelSet;

        /// <summary>
        /// 웨이브 출력
        /// </summary>
        private WaveOut waveOut;

        /// <summary>
        /// 활성 웨이브 스트림
        /// </summary>
        private WaveStream activeWaveStream;

        /// <summary>
        /// 웨이브 채널 32
        /// </summary>
        private WaveChannel32 waveChannel32;

        /// <summary>
        /// 샘플 수집기
        /// </summary>
        private SampleAggregator sampleAggregator;

        /// <summary>
        /// 웨이브폼 샘플 수집기
        /// </summary>
        private SampleAggregator waveformSampleAggregator;
        
        /// <summary>
        /// 보류 웨이브폼 경로
        /// </summary>
        private string pendingWaveformPath;

        /// <summary>
        /// 전체 레벨 단정도 실수 배열
        /// </summary>
        private float[] fullLevelFloatArray;

        /// <summary>
        /// 웨이브폼 단정도 실수 배열
        /// </summary>
        private float[] waveformFloatArray;

        private TagLib.File fileTag;

        /// <summary>
        /// 반복 시작 시간 범위
        /// </summary>
        private TimeSpan repeatStartTimeSpan;

        /// <summary>
        /// 반복 중단 시간 범위
        /// </summary>
        private TimeSpan repeatStopTimeSpan;

        /// <summary>
        /// 반복 설정 중 여부
        /// </summary>
        private bool inRepeatSet;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 인스턴스 - Instance

        /// <summary>
        /// 인스턴스
        /// </summary>
        public static NAudioEngine Instance
        {
            get
            {
                if(_instance == null)
                {
                    _instance = new NAudioEngine();
                }

                return _instance;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 파일 태그 - FileTag

        /// <summary>
        /// 파일 태그
        /// </summary>
        public TagLib.File FileTag
        {
            get
            {
                return this.fileTag;
            }
            set
            {
                TagLib.File previousValue = fileTag;

                this.fileTag = value;

                if(previousValue != this.fileTag)
                {
                    FirePropertyChangedEvent("FileTag");
                }
            }
        }

        #endregion
        #region 활성 웨이브 스트림 - ActiveWaveStream

        /// <summary>
        /// 활성 웨이브 스트림
        /// </summary>
        public WaveStream ActiveWaveStream
        {
            get
            {
                return this.activeWaveStream;
            }
            protected set
            {
                WaveStream previousValue = this.activeWaveStream;

                this.activeWaveStream = value;

                if(previousValue != this.activeWaveStream)
                {
                    FirePropertyChangedEvent("ActiveStream");
                }
            }
        }

        #endregion
        #region 재생 가능 여부 - CanPlay

        /// <summary>
        /// 재생 가능 여부
        /// </summary>
        public bool CanPlay
        {
            get
            {
                return this.canPlay;
            }
            protected set
            {
                bool previousValue = this.canPlay;

                this.canPlay = value;

                if(previousValue != this.canPlay)
                {
                    FirePropertyChangedEvent("CanPlay");
                }
            }
        }

        #endregion
        #region 중지 가능 여부 - CanPause

        /// <summary>
        /// 중지 가능 여부
        /// </summary>
        public bool CanPause
        {
            get
            {
                return this.canPause;
            }
            protected set
            {
                bool previousValue = this.canPause;

                this.canPause = value;

                if(previousValue != this.canPause)
                {
                    FirePropertyChangedEvent("CanPause");
                }
            }
        }

        #endregion
        #region 중단 가능 여부 - CanStop

        /// <summary>
        /// 중단 가능 여부
        /// </summary>
        public bool CanStop
        {
            get
            {
                return this.canStop;
            }
            protected set
            {
                bool previousValue = this.canStop;

                this.canStop = value;

                if(previousValue != this.canStop)
                {
                    FirePropertyChangedEvent("CanStop");
                }
            }
        }

        #endregion
        #region 재생 여부 - IsPlaying

        /// <summary>
        /// 재생 여부
        /// </summary>
        public bool IsPlaying
        {
            get
            {
                return this.isPlaying;
            }
            protected set
            {
                bool previousValue = this.isPlaying;

                this.isPlaying = value;

                if(previousValue != this.isPlaying)
                {
                    FirePropertyChangedEvent("IsPlaying");
                }

                this.positionDispatcherTimer.IsEnabled = value;
            }
        }    

        #endregion

        #region (IWaveformPlayer) 선택 시작 - SelectionBegin

        /// <summary>
        /// 선택 시작
        /// </summary>
        public TimeSpan SelectionBegin
        {
            get
            {
                return this.repeatStartTimeSpan;
            }
            set
            {
                if(!this.inRepeatSet)
                {
                    this.inRepeatSet = true;

                    TimeSpan previousTimeSpan = this.repeatStartTimeSpan;

                    this.repeatStartTimeSpan = value;

                    if(previousTimeSpan != this.repeatStartTimeSpan)
                    {
                        FirePropertyChangedEvent("SelectionBegin");
                    }

                    this.inRepeatSet = false;
                }
            }
        }

        #endregion
        #region (IWaveformPlayer) 선택 종료 - SelectionEnd

        /// <summary>
        /// 선택 종료
        /// </summary>
        public TimeSpan SelectionEnd
        {
            get
            {
                return this.repeatStopTimeSpan;
            }
            set
            {
                if(!this.inChannelSet)
                {
                    this.inRepeatSet = true;

                    TimeSpan previousTimeSpan = this.repeatStopTimeSpan;

                    this.repeatStopTimeSpan = value;

                    if(previousTimeSpan != this.repeatStopTimeSpan)
                    {
                        FirePropertyChangedEvent("SelectionEnd");
                    }

                    this.inRepeatSet = false;
                }
            }
        }

        #endregion
        #region (IWaveformPlayer) 웨이브폼 단정도 실수 배열 - WaveFormFloatArray

        /// <summary>
        /// 웨이브폼 단정도 실수 배열
        /// </summary>
        public float[] WaveFormFloatArray
        {
            get
            {
                return this.waveformFloatArray;
            }
            protected set
            {
                float[] previousFloatArray = this.waveformFloatArray;

                this.waveformFloatArray = value;

                if(previousFloatArray != this.waveformFloatArray)
                {
                    FirePropertyChangedEvent("WaveformData");
                }
            }
        }

        #endregion
        #region (IWaveformPlayer) 채널 길이 - ChannelLength

        /// <summary>
        /// 채널 길이
        /// </summary>
        public double ChannelLength
        {
            get
            {
                return this.channelLength;
            }
            protected set
            {
                double previousValue = this.channelLength;

                this.channelLength = value;

                if(previousValue != this.channelLength)
                {
                    FirePropertyChangedEvent("ChannelLength");
                }
            }
        }

        #endregion
        #region (IWaveformPlayer) 채널 위치 - ChannelPosition

        /// <summary>
        /// 채널 위치
        /// </summary>
        public double ChannelPosition
        {
            get
            {
                return this.channelPosition;
            }
            set
            {
                if(!this.inChannelSet)
                {
                    this.inChannelSet = true;

                    double previousValue = this.channelPosition;

                    double position = Math.Max(0, Math.Min(value, ChannelLength));

                    if(!this.inChannelTimerUpdate && ActiveWaveStream != null)
                    {
                        ActiveWaveStream.Position = (long)((position / ActiveWaveStream.TotalTime.TotalSeconds) * ActiveWaveStream.Length);
                    }

                    this.channelPosition = position;

                    if(previousValue != this.channelPosition)
                    {
                        FirePropertyChangedEvent("ChannelPosition");
                    }

                    this.inChannelSet = false;
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 생성자 - NAudioEngine()

        /// <summary>
        /// 생성자
        /// </summary>
        private NAudioEngine()
        {
            this.positionDispatcherTimer.Interval = TimeSpan.FromMilliseconds(50);

            this.waveformGenerateBackgroundWorker.WorkerSupportsCancellation = true;


            this.positionDispatcherTimer.Tick                        += positionDispatcherTimer_Tick;
            this.waveformGenerateBackgroundWorker.DoWork             += waveformGenerateBackgroundWorker_DoWork;
            this.waveformGenerateBackgroundWorker.RunWorkerCompleted += waveformGenerateBackgroundWorker_RunWorkerCompleted;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파일 열기 - OpenFile(path)

        /// <summary>
        /// 파일 열기
        /// </summary>
        /// <param name="path">경로</param>
        public void OpenFile(string path)
        {
            Stop();

            if(ActiveWaveStream != null)
            {
                SelectionBegin  = TimeSpan.Zero;
                SelectionEnd    = TimeSpan.Zero;
                ChannelPosition = 0;
            }
            
            StopAndCloseStream();            

            if(File.Exists(path))
            {
                try
                {
                    this.waveOut = new WaveOut()
                    {
                        DesiredLatency = 100
                    };

                    ActiveWaveStream = new Mp3FileReader(path);

                    this.waveChannel32 = new WaveChannel32(ActiveWaveStream);

                    this.sampleAggregator = new SampleAggregator(this.fftDataSize);

                    this.waveChannel32.Sample += waveChannel32_Sample1;

                    this.waveOut.Init(this.waveChannel32);

                    ChannelLength = this.waveChannel32.TotalTime.TotalSeconds;

                    FileTag = TagLib.File.Create(path);

                    GenerateWaveformData(path);

                    CanPlay = true;
                }
                catch
                {
                    ActiveWaveStream = null;

                    CanPlay = false;
                }
            }
        }

        #endregion
        #region 재생하기 - Play()

        /// <summary>
        /// 재생하기
        /// </summary>
        public void Play()
        {
            if(CanPlay)
            {
                this.waveOut.Play();

                IsPlaying = true;
                CanPause  = true;
                CanPlay   = false;
                CanStop   = true;
            }
        }

        #endregion
        #region 중지하기 - Pause()

        /// <summary>
        /// 중지하기
        /// </summary>
        public void Pause()
        {
            if(IsPlaying && CanPause)
            {
                this.waveOut.Pause();

                IsPlaying = false;
                CanPlay   = true;
                CanPause  = false;
            }
        }

        #endregion
        #region 중단하기 - Stop()

        /// <summary>
        /// 중단하기
        /// </summary>
        public void Stop()
        {
            if(this.waveOut != null)
            {
                this.waveOut.Stop();
            }

            IsPlaying = false;
            CanStop   = false;
            CanPlay   = true;
            CanPause  = false;
        }

        #endregion
 
        #region (ISpectrumPlayer) FFT 데이터 설정하기 - SetFFTData(targetArray)

        /// <summary>
        /// FFT 데이터 설정하기
        /// </summary>
        /// <param name="targetArray">타겟 배열</param>
        /// <returns>처리 결과</returns>
        public bool SetFFTData(float[] targetArray)
        {
            this.sampleAggregator.SetFFTData(targetArray);

            return this.isPlaying;
        }

        #endregion
        #region (ISpectrumPlayer) FFT 주기 인덱스 구하기 - GetFFTFrequencyIndex(frequency)

        /// <summary>
        /// FFT 주기 인덱스 구하기
        /// </summary>
        /// <param name="frequency">주기</param>
        /// <returns>FFT 주기 인덱스</returns>
        public int GetFFTFrequencyIndex(int frequency)
        {
            double maximumFrequency;

            if(ActiveWaveStream != null)
            {
                maximumFrequency = ActiveWaveStream.WaveFormat.SampleRate / 2.0d;
            }
            else
            {
                maximumFrequency = 22050; // 디폴트 44.1 kHz 샘플 비율을 가정한다.
            }

            return (int)((frequency / maximumFrequency) * (this.fftDataSize / 2));
        }

        #endregion

        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            
            GC.SuppressFinalize(this);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 리소스 해제하기 - Dispose(disposing)

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        /// <param name="disposing">해제 여부</param>
        protected virtual void Dispose(bool disposing)
        {
            if(!this.disposed)
            {
                if(disposing)
                {
                    StopAndCloseStream();
                }

                this.disposed = true;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 위치 디스패처 타이머 틱 처리하기 - positionDispatcherTimer_Tick(sender, e)

        /// <summary>
        /// 위치 디스패처 타이머 틱 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void positionDispatcherTimer_Tick(object sender, EventArgs e)
        {
            this.inChannelTimerUpdate = true;

            ChannelPosition = ((double)ActiveWaveStream.Position / (double)ActiveWaveStream.Length) * ActiveWaveStream.TotalTime.TotalSeconds;

            this.inChannelTimerUpdate = false;
        }

        #endregion
        #region 웨이브폼 생성 백그라운드 작업자 작업시 처리하기 - waveformGenerateBackgroundWorker_DoWork(sender, e)

        /// <summary>
        /// 웨이브폼 생성 백그라운드 작업자 작업시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void waveformGenerateBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            WaveformGenerationParameter waveformGenerationParameter = e.Argument as WaveformGenerationParameter;

            Mp3FileReader mp3FileReader = new Mp3FileReader(waveformGenerationParameter.Path);

            WaveChannel32 waveChannel32 = new WaveChannel32(mp3FileReader);

            waveChannel32.Sample += waveChannel32_Sample2;
            
            int    frameLength    = this.fftDataSize;
            int    frameCount     = (int)((double)waveChannel32.Length / (double)frameLength);
            int    waveformLength = frameCount * 2;
            byte[] readByteArray  = new byte[frameLength];

            this.waveformSampleAggregator = new SampleAggregator(frameLength);
   
            float maximumLeftPointLevel  = float.MinValue;
            float maximumRightPointLevel = float.MinValue;

            int currentPointIndex = 0;

            float[] waveformCompressedPointArray = new float[waveformGenerationParameter.PointCount];

            List<float> waveformList              = new List<float>();
            List<int>   waveMaximumPointIndexList = new List<int>();
            
            for(int i = 1; i <= waveformGenerationParameter.PointCount; i++)
            {
                waveMaximumPointIndexList.Add
                (
                    (int)Math.Round
                    (
                        waveformLength * ((double)i / (double)waveformGenerationParameter.PointCount),
                        0
                    )
                );
            }

            int readCount = 0;

            while(currentPointIndex * 2 < waveformGenerationParameter.PointCount)
            {
                waveChannel32.Read(readByteArray, 0, readByteArray.Length);

                waveformList.Add(this.waveformSampleAggregator.LeftMaximumVolume );
                waveformList.Add(this.waveformSampleAggregator.RightMaximumVolume);

                if(this.waveformSampleAggregator.LeftMaximumVolume > maximumLeftPointLevel)
                {
                    maximumLeftPointLevel = this.waveformSampleAggregator.LeftMaximumVolume;
                }

                if(this.waveformSampleAggregator.RightMaximumVolume > maximumRightPointLevel)
                {
                    maximumRightPointLevel = this.waveformSampleAggregator.RightMaximumVolume;
                }

                if(readCount > waveMaximumPointIndexList[currentPointIndex])
                {
                    waveformCompressedPointArray[currentPointIndex * 2    ] = maximumLeftPointLevel;
                    waveformCompressedPointArray[currentPointIndex * 2 + 1] = maximumRightPointLevel;

                    maximumLeftPointLevel  = float.MinValue;
                    maximumRightPointLevel = float.MinValue;

                    currentPointIndex++;
                }

                if(readCount % 3000 == 0)
                {
                    float[] clonedData = (float[])waveformCompressedPointArray.Clone();

                    MainApplication.Current.Dispatcher.Invoke
                    (
                        new Action
                        (
                            () => { WaveFormFloatArray = clonedData; }
                        )
                    );
                }

                if(this.waveformGenerateBackgroundWorker.CancellationPending)
                {
                    e.Cancel = true;

                    break;
                }

                readCount++;
            }

            float[] finalFloatArray = (float[])waveformCompressedPointArray.Clone();

            MainApplication.Current.Dispatcher.Invoke
            (
                new Action
                (
                    () => {

                        this.fullLevelFloatArray = waveformList.ToArray();

                        WaveFormFloatArray = finalFloatArray;
                    }
                )
            );

            waveChannel32.Close();

            waveChannel32.Dispose();

            waveChannel32 = null;

            mp3FileReader.Close();

            mp3FileReader.Dispose();

            mp3FileReader = null;
        }

        #endregion
        #region 웨이브폼 생성 백그라운드 작업자 실행 작업자 완료시 처리하기 - waveformGenerateBackgroundWorker_RunWorkerCompleted(sender, e)

        /// <summary>
        /// 웨이브폼 생성 백그라운드 작업자 실행 작업자 완료시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void waveformGenerateBackgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if(e.Cancelled)
            {
                if(!this.waveformGenerateBackgroundWorker.IsBusy && WAVEFORM_COMPRESSED_POINT_COUNT != 0)
                {
                    this.waveformGenerateBackgroundWorker.RunWorkerAsync
                    (
                        new WaveformGenerationParameter
                        (
                            WAVEFORM_COMPRESSED_POINT_COUNT,
                            this.pendingWaveformPath
                        )
                    );
                }
            }
        }

        #endregion
        #region 웨이브 채널 32 샘플 처리하기 1 - waveChannel32_Sample1(sender, e)

        /// <summary>
        /// 웨이브 채널 32 샘플 처리하기 1
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void waveChannel32_Sample1(object sender, SampleEventArgs e)
        {
            this.sampleAggregator.Add(e.Left, e.Right);

            long repeatStartPosition = (long)((SelectionBegin.TotalSeconds / ActiveWaveStream.TotalTime.TotalSeconds) * ActiveWaveStream.Length);
            long repeatStopPosition  = (long)((SelectionEnd.TotalSeconds   / ActiveWaveStream.TotalTime.TotalSeconds) * ActiveWaveStream.Length);

            if(((SelectionEnd - SelectionBegin) >= TimeSpan.FromMilliseconds(REPEAT_THRESHOLD)) && ActiveWaveStream.Position >= repeatStopPosition)
            {
                this.sampleAggregator.Clear();

                ActiveWaveStream.Position = repeatStartPosition;
            }
        }

        #endregion
        #region 웨이브 채널 32 샘플 처리하기 2 - waveChannel32_Sample2(sender, e)

        /// <summary>
        /// 웨이브 채널 32 샘플 처리하기 2
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void waveChannel32_Sample2(object sender, SampleEventArgs e)
        {
            this.waveformSampleAggregator.Add(e.Left, e.Right);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(propertyName)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="propertyName">속성명</param>
        private void FirePropertyChangedEvent(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion
        #region 웨이브폼 데이터 생성하기 - GenerateWaveformData(path)

        /// <summary>
        /// 웨이브폼 데이터 생성하기
        /// </summary>
        /// <param name="path">경로</param>
        private void GenerateWaveformData(string path)
        {
            if(this.waveformGenerateBackgroundWorker.IsBusy)
            {
                this.pendingWaveformPath = path;

                this.waveformGenerateBackgroundWorker.CancelAsync();

                return;
            }

            if(!this.waveformGenerateBackgroundWorker.IsBusy && WAVEFORM_COMPRESSED_POINT_COUNT != 0)
            {
                this.waveformGenerateBackgroundWorker.RunWorkerAsync
                (
                    new WaveformGenerationParameter
                    (
                        WAVEFORM_COMPRESSED_POINT_COUNT,
                        path
                    )
                );
            }
        }

        #endregion
        #region 스트림 중단하고 닫기 - StopAndCloseStream()

        /// <summary>
        /// 스트림 중단하고 닫기
        /// </summary>
        private void StopAndCloseStream()
        {
            if(this.waveOut != null)
            {
                this.waveOut.Stop();
            }

            if(this.activeWaveStream != null)
            {
                this.waveChannel32.Close();

                this.waveChannel32 = null;

                ActiveWaveStream.Close();

                ActiveWaveStream = null;
            }

            if(this.waveOut != null)
            {
                this.waveOut.Dispose();

                this.waveOut = null;
            }
        }

        #endregion
    }
}